from abc import ABC
from datetime import datetime, timezone
from typing_extensions import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Optional,
    Self,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID, uuid4

from pydantic import BaseModel, Field
from syft_core import Client as SyftBoxClient

from syft_rds.display_utils.formatter import (
    ANSIPydanticFormatter,
    PydanticFormatter,
)

if TYPE_CHECKING:
    from syft_rds.client.rds_client import RDSClient


def _utcnow():
    return datetime.now(tz=timezone.utc)


class ItemBase(BaseModel, ABC):
    __schema_name__: str
    __display_formatter__: ClassVar[PydanticFormatter] = ANSIPydanticFormatter()

    uid: UUID = Field(default_factory=uuid4)
    created_by: str | None = None
    created_at: datetime = Field(default_factory=_utcnow)
    updated_at: datetime = Field(default_factory=_utcnow)
    client_id: UUID | None = None

    class Config:
        arbitrary_types_allowed: bool = True

    def register_client(self, client: "RDSClient") -> Self:
        self._register_client_id_recursive(client.uid)
        return self

    def _register_client_id_recursive(self, client_id: UUID) -> Self:
        self.client_id = client_id
        for field in self.model_fields.keys():
            field_value = getattr(self, field)
            if isinstance(field_value, ItemBase):
                field_value._register_client_id_recursive(client_id)
        return self

    @property
    def _client(self) -> "RDSClient":
        from syft_rds.client.client_registry import GlobalClientRegistry

        if self.client_id is None:
            raise ValueError("Client ID not set")
        return GlobalClientRegistry.get_client(self.client_id)

    @property
    def _syftbox_client(self) -> SyftBoxClient:
        return self._client._syftbox_client

    @classmethod
    def type_name(cls) -> str:
        return cls.__name__.lower()

    def refresh(self, in_place: bool = True) -> Self:
        new_item = self._client.for_type(type(self)).get(self.uid)
        return self.apply_update(new_item, in_place=in_place)

    def apply_update(
        self, other: Union[Self, "ItemBaseUpdate[Self]"], in_place: bool = True
    ) -> Self:
        """Updates this instance with the provided update instance."""
        if other.uid != self.uid:
            raise ValueError(
                f"Cannot apply update with UID {other.uid} to instance with UID {self.uid}"
            )
        if isinstance(other, type(self)):
            update_dict = other.model_dump()
        elif isinstance(other, ItemBaseUpdate):
            update_target_type = other.get_target_model()
            if other.get_target_model() is not type(self):
                raise ValueError(
                    f"Attempted to apply update for {update_target_type} to {type(self)}"
                )
            update_dict = other.model_dump(exclude_unset=True)
            update_dict["updated_at"] = _utcnow()
        else:
            raise TypeError(
                f"Cannot apply update of type {type(other)} to {type(self)}"
            )

        if in_place:
            for field_name, value in update_dict.items():
                if field_name in self.model_fields:
                    setattr(self, field_name, value)
            return self
        else:
            return self.model_copy(update=update_dict)

    def __str__(self) -> str:
        return self.__display_formatter__.format_str(self)

    def __repr__(self) -> str:
        return self.__display_formatter__.format_repr(self)

    def _repr_html_(self) -> str:
        return self.__display_formatter__.format_html(self)

    def _repr_markdown_(self) -> str:
        return self.__display_formatter__.format_markdown(self)


T = TypeVar("T", bound=ItemBase)


class ItemBaseCreate(BaseModel, Generic[T]):
    @classmethod
    def get_target_model(cls) -> Type[T]:
        return cls.__bases__[0].__pydantic_generic_metadata__["args"][0]  # type: ignore

    def to_item(self, extra: Optional[dict[str, Any]] = None) -> T:
        model_cls = self.get_target_model()
        extra = extra or {}
        return model_cls(**self.model_dump(), **extra)


class ItemBaseUpdate(BaseModel, Generic[T]):
    uid: UUID

    @classmethod
    def get_target_model(cls) -> Type[T]:
        return cls.__bases__[0].__pydantic_generic_metadata__["args"][0]  # type: ignore



import shutil
import sys
import tempfile
from pathlib import Path
from typing import Optional

from loguru import logger
from syft_core import Client as SyftBoxClient
from syft_core import SyftClientConfig

from syft_rds.client.rds_client import init_session as init_session_rds
from syft_rds.client.utils import PathLike
from syft_rds.server.app import create_app


def setup_logger(level: str = "DEBUG") -> None:
    """
    Setup loguru logger with custom filtering.

    Args:
        level (str): Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
    """
    # Remove default handler
    logger.remove()

    # Add custom handler that filters out noisy logs
    logger.add(
        sys.stderr,
        level=level,
        filter=lambda record: "syft_event.server2" not in record["name"],
    )


class RDSStack:
    """
    Simple wrapper for RDS stack with SyftBox clients and RDS server
    Includes 1 DO and 1 DS client
    """

    def __init__(
        self,
        do_client: SyftBoxClient,
        ds_client: SyftBoxClient,
        **config_kwargs,
    ):
        self.do_client = do_client
        self.ds_client = ds_client

        self.server = create_app(do_client)
        self.server.start()

        self.do_rds_client = init_session_rds(
            host=do_client.email,
            syftbox_client=do_client,
            **config_kwargs,
        )

        self.ds_rds_client = init_session_rds(
            host=do_client.email,
            syftbox_client=ds_client,
            **config_kwargs,
        )

    def stop(self) -> None:
        return self.server.stop()


def _prepare_root_dir(
    root_dir: Optional[PathLike] = None,
    reset: bool = False,
    key: str = "shared_client_dir",
) -> Path:
    if root_dir is None:
        root_path = Path(tempfile.gettempdir(), key)
    else:
        root_path = Path(root_dir).resolve() / key

    if reset and root_path.is_dir():
        try:
            shutil.rmtree(root_path)
        except Exception as e:
            logger.warning(f"Failed to reset directory {root_path}: {e}")

    root_path.mkdir(parents=True, exist_ok=True)
    return root_path


def remove_rds_stack_dir(
    key: str = "shared_client_dir", root_dir: Optional[PathLike] = None
) -> None:
    root_path = (
        Path(root_dir).resolve() / key if root_dir else Path(tempfile.gettempdir(), key)
    )

    if not root_path.exists():
        logger.warning(f"⚠️ Skipping removal, as path {root_path} does not exist")
        return None

    try:
        shutil.rmtree(root_path)
        logger.info(f"✅ Successfully removed directory {root_path}")
    except Exception as e:
        logger.error(f"❌ Failed to remove directory {root_path}: {e}")


def setup_rds_stack(
    root_dir: Optional[PathLike] = None,
    do_email: str = "data_owner@test.openmined.org",
    ds_email: str = "data_scientist@test.openmined.org",
    reset: bool = False,
    log_level: str = "DEBUG",
    key: str = "shared_client_dir",
    **config_kwargs,
) -> RDSStack:
    setup_logger(level=log_level)
    root_dir = _prepare_root_dir(root_dir, reset, key)

    logger.warning(
        "Using shared data directory for both clients. "
        "Any file permission checks will be skipped as both clients have access to the same files."
    )

    # We also save the config files in the root dir
    do_client_config = SyftClientConfig(
        email=do_email,
        server_url="http://localhost:8080",  # Explicit server_url for proper bootstrap
        client_url="http://localhost:5000",  # not used, just for local dev
        path=root_dir / "do_config.json",
        data_dir=root_dir,
    ).save()
    do_client = SyftBoxClient(
        do_client_config,
    )

    ds_client_config = SyftClientConfig(
        email=ds_email,
        server_url="http://localhost:8080",  # Explicit server_url for proper bootstrap
        client_url="http://localhost:5001",  # not used, just for local dev
        path=root_dir / "ds_config.json",
        data_dir=root_dir,
    ).save()
    ds_client = SyftBoxClient(ds_client_config)

    logger.info(f"Launching mock RDS stack in {root_dir}")

    return RDSStack(
        do_client=do_client,
        ds_client=ds_client,
        **config_kwargs,
    )


class SingleRDSStack:
    """A single RDS stack with a SyftBox client and a RDS server"""

    def __init__(self, client: SyftBoxClient, **config_kwargs):
        self.client = client
        self.server = create_app(client)
        self.server.start()

    def init_session(self, host, **config_kwargs):
        return init_session_rds(
            host=host,
            syftbox_client=self.client,
            **config_kwargs,
        )

    def stop(self) -> None:
        return self.server.stop()


def _get_syftbox_client(
    email: str,
    root_dir: PathLike,
) -> SyftBoxClient:
    """
    Get a SyftBox client for testing.

    Args:
        email (str): Email address of the user.
        root_dir (PathLike): Directory to store the client files.

    Returns:
        SyftBoxClient: The SyftBox client.
    """
    # We also save the config files in the root dir
    client_config = SyftClientConfig(
        email=email,
        server_url="http://localhost:8080",  # Explicit server_url for proper bootstrap
        client_url="http://localhost:5000",  # not used, just for local dev
        path=root_dir / f"{email}.config.json",
        data_dir=root_dir,
    ).save()
    return SyftBoxClient(client_config)


def setup_rds_server(
    email: str,
    root_dir: Optional[PathLike] = None,
    reset: bool = False,
    key: str = "shared_client_dir",
    log_level: str = "DEBUG",
    **config_kwargs,
):
    """
    Setup a mock RDS server for testing.

    Args:
        email (str): Email address of the user.
        root_dir (Optional[PathLike]): Directory to store the server files.
        reset (bool): Whether to reset the directory.
        key (str): Key for the directory.
        log_level (str): Log level for logging.
        **config_kwargs: Additional configuration arguments.

    Returns:
        RDSStack: The RDS stack with the server and clients.
    """
    setup_logger(level=log_level)
    root_dir = _prepare_root_dir(root_dir, reset, key)

    client = _get_syftbox_client(email=email, root_dir=root_dir)

    logger.info(f"Launching mock RDS server in {root_dir.resolve()}")

    return SingleRDSStack(
        client=client,
        **config_kwargs,
    )



from typing import TYPE_CHECKING, Type, TypeVar

from syft_core import Client as SyftBoxClient

from syft_rds.client.local_stores.base import CRUDLocalStore
from syft_rds.client.local_stores.custom_function import CustomFunctionLocalStore
from syft_rds.client.local_stores.dataset import DatasetLocalStore
from syft_rds.client.local_stores.job import JobLocalStore
from syft_rds.client.local_stores.runtime import RuntimeLocalStore
from syft_rds.client.local_stores.user_code import UserCodeLocalStore
from syft_rds.models import Dataset, Job, Runtime, UserCode
from syft_rds.models.base import ItemBase, ItemBaseCreate, ItemBaseUpdate
from syft_rds.models.custom_function_models import CustomFunction

if TYPE_CHECKING:
    from syft_rds.client.rds_client import RDSClientConfig

T = TypeVar("T", bound=ItemBase)


class LocalStore:
    def __init__(self, config: "RDSClientConfig", syftbox_client: SyftBoxClient):
        self.config = config
        self.syftbox_client = syftbox_client
        self.job = JobLocalStore(self.config, self.syftbox_client)
        self.user_code = UserCodeLocalStore(self.config, self.syftbox_client)
        self.runtime = RuntimeLocalStore(self.config, self.syftbox_client)
        self.dataset = DatasetLocalStore(self.config, self.syftbox_client)
        self.custom_function = CustomFunctionLocalStore(
            self.config, self.syftbox_client
        )

        self._type_map = {
            Job: self.job,
            UserCode: self.user_code,
            Runtime: self.runtime,
            Dataset: self.dataset,
            CustomFunction: self.custom_function,
        }

    def for_type(
        self, type_: Type[T]
    ) -> CRUDLocalStore[T, ItemBaseCreate, ItemBaseUpdate]:
        if type_ not in self._type_map:
            raise ValueError(f"No local store found for type {type_}")
        return self._type_map[type_]



from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Literal, Optional, Type
from uuid import UUID, uuid4

from pydantic import BaseModel, Field
from syft_core import Client as SyftBoxClient

from syft_rds.client.local_store import LocalStore
from syft_rds.client.rpc import RPCClient, T
from syft_rds.models import GetAllRequest, GetOneRequest, Job, Runtime

if TYPE_CHECKING:
    from syft_rds.client.rds_client import RDSClient


class ClientRunnerConfig(BaseModel):
    runtime: Runtime | None = None
    timeout: int = 60
    job_output_folder: Path = Field(
        default_factory=lambda: Path(".server/syft-rds-jobs")
    )


class RDSClientConfig(BaseModel):
    uid: UUID = Field(
        default_factory=uuid4
    )  # used to register the client & reference it from other objects
    host: str
    app_name: str = "RDS"
    rpc_expiry: str = "5m"
    runner_config: ClientRunnerConfig = Field(default_factory=ClientRunnerConfig)


class RDSClientBase:
    def __init__(
        self, config: RDSClientConfig, rpc_client: RPCClient, local_store: LocalStore
    ) -> None:
        self.config = config
        self.rpc = rpc_client
        self.local_store = local_store

    @property
    def host(self) -> str:
        return self.config.host

    @property
    def _syftbox_client(self) -> SyftBoxClient:
        return self.rpc.connection.sender_client

    @property
    def email(self) -> str:
        return self._syftbox_client.email

    @property
    def is_admin(self) -> bool:
        return self.host == self.email


class RDSClientModule(RDSClientBase, Generic[T]):
    ITEM_TYPE: ClassVar[Type[T]]

    def __init__(
        self,
        config: RDSClientConfig,
        rpc_client: RPCClient,
        local_store: LocalStore,
        parent: "Optional[RDSClient]" = None,
    ) -> None:
        super().__init__(config, rpc_client, local_store)
        self.parent = parent

    @property
    def rds(self) -> "RDSClient":
        """
        Returns the parent RDSClient, raises an error if not set.
        Used for accessing other client modules from this module, e.g. JobRDSClient().rds.dataset.get_all() -> DatasetRDSClient
        """
        if self.parent is None:
            raise ValueError("Parent client not set")
        return self.parent

    def get_all(
        self,
        order_by: str = "created_at",
        sort_order: str = "desc",
        limit: Optional[int] = None,
        offset: int = 0,
        mode: Literal["local", "rpc"] = "local",
        **filters: Any,
    ) -> list[Job]:
        req = GetAllRequest(
            order_by=order_by,
            sort_order=sort_order,
            limit=limit,
            offset=offset,
            filters=filters,
        )

        if mode == "local":
            return self.local_store.for_type(self.ITEM_TYPE).get_all(req)
        elif mode == "rpc":
            return self.rpc.for_type(self.ITEM_TYPE).get_all(req)
        else:
            raise ValueError(f"Invalid mode {mode}")

    def get(
        self,
        uid: Optional[UUID] = None,
        mode: Literal["local", "rpc"] = "local",
        **filters: Any,
    ) -> T:
        req = GetOneRequest(uid=uid, filters=filters)
        if mode == "local":
            return self.local_store.for_type(self.ITEM_TYPE).get_one(req)
        elif mode == "rpc":
            return self.rpc.for_type(self.ITEM_TYPE).get_one(req)
        else:
            raise ValueError(f"Invalid mode {mode}")



from pathlib import Path
from typing import Type

from syft_rds.models.base import ItemBase

PUBLIC_FILES_PERMISSIONS = """
rules:
- pattern: '**'
  access:
    read:
    - '*'
"""


class PublicFileService:
    def __init__(self, app_dir: Path):
        """Service for managing files for the RDS app that everyone can read.

        General structure:
        APP_FILES_DIR/
        ├── syft.pub.yaml  # Read permissions for everyone
        ├── {item_type}/
        │   ├── {item_uid}/
        │   │   ├── public_file.txt

        Args:
            app_dir: The root application directory
        """
        self.app_dir = app_dir
        self.public_files_dir = app_dir / "public_files"
        self._init_public_files_dir()

    def _init_public_files_dir(self) -> None:
        self.public_files_dir.mkdir(exist_ok=True)
        perm_path = self.public_files_dir / "syft.pub.yaml"
        perm_path.write_text(PUBLIC_FILES_PERMISSIONS)

    def dir_for_type(self, type_: Type[ItemBase]) -> Path:
        if not issubclass(type_, ItemBase):
            raise ValueError(f"Type {type_} must be a subclass of ItemBase")
        type_dir = self.public_files_dir / type_.__name__
        type_dir.mkdir(exist_ok=True)
        return type_dir

    def dir_for_item(self, item: ItemBase) -> Path:
        type_dir = self.dir_for_type(type(item))
        item_dir = type_dir / str(item.uid)
        item_dir.mkdir(exist_ok=True)
        return item_dir



from pathlib import Path
from typing import Type

from syft_rds.models.base import ItemBase

USER_FILES_DIR = "user_files"

USER_FILES_PERMISSION_TEMPLATE = """
rules:
- pattern: '**'
  access:
    read:
    - '{useremail}'
"""


class UserFileService:
    """Service for managing user file directories with proper permissions.
    Users have read access to all files in their directory, and no access to other users' files.

    General structure:
    USER_FILES_DIR/
    ├── {useremail}/
    │   ├── syft.pub.yaml  # Read permissions for the user {useremail}
    │   ├── {item_type}/
    │   │   ├── {item_uid}/

    Example: job outputs for uid "1234" for user 'alice@openmined.org' go in:
    USER_FILES_DIR/alice@openmined.org/Job/1234/

    To get the path for a specific user/job:
    user_file_service.dir_for_item(user="alice@openmined.org", item=job)
    """

    def __init__(self, app_dir: Path):
        """Initialize the user file service.

        Args:
            app_dir: The root application directory
        """
        self.app_dir = app_dir
        self.user_files_dir = app_dir / USER_FILES_DIR
        self._init_user_files_dir()

    def _init_user_files_dir(self) -> None:
        """Initialize the user files directory with proper permissions."""
        self.user_files_dir.mkdir(exist_ok=True)

    def _is_valid_dirname(self, user: str) -> str:
        if not user or user in {".", ".."}:
            raise ValueError("Invalid directory name: cannot be '.' or '..' or empty")
        disallowed = {"/", "\\", "\x00"}
        if any(char in user for char in disallowed):
            raise ValueError(f"Invalid directory name: contains one of {disallowed}")
        return user

    def dir_for_user(self, user: str) -> Path:
        """Get the user's file directory, creating it if it doesn't exist"""
        user = self._is_valid_dirname(user)
        user_dir = self.user_files_dir / user
        user_dir.mkdir(exist_ok=True, parents=True)
        perm_file_path = user_dir / "syft.pub.yaml"
        if not perm_file_path.exists():
            with perm_file_path.open("w") as perm_file:
                perm_file.write(USER_FILES_PERMISSION_TEMPLATE.format(useremail=user))
        return user_dir

    def dir_for_type(self, user: str, type_: Type[ItemBase]) -> Path:
        """Get the user's directory for a specific item type"""
        user_dir = self.dir_for_user(user)
        item_dir = user_dir / type_.__name__
        item_dir.mkdir(exist_ok=True)
        return item_dir

    def dir_for_item(self, user: str, item: ItemBase) -> Path:
        """Get the directory for a specific item instance"""
        type_dir = self.dir_for_type(user, type(item))
        item_dir = type_dir / str(item.uid)
        item_dir.mkdir(exist_ok=True)
        return item_dir



import sys
from abc import ABC, abstractmethod
from typing import Any, ClassVar

from pydantic import BaseModel


class PydanticFormatter(ABC):
    """Interface for Pydantic model formatters, to be used with PydanticFormatterMixin"""

    @abstractmethod
    def format_str(self, model: BaseModel) -> str: ...

    @abstractmethod
    def format_repr(self, model: BaseModel) -> str: ...

    def format_markdown(self, model: BaseModel) -> str | None:
        return None

    def format_html(self, model: BaseModel) -> str | None:
        return None


class ANSIPydanticFormatter(PydanticFormatter):
    """Format Pydantic models multiline string with ANSI colors"""

    def __init__(self):
        self.use_colors = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()

    def format_class_name(self, name: str) -> str:
        if self.use_colors:
            return f"\033[1;36m{name}\033[0m"
        return name

    def format_field(self, key: str, value: Any) -> str:
        value_str = str(value)
        formatted_key = key
        formatted_value = value_str
        if self.use_colors:
            formatted_key = f"\033[1m{key}\033[0m"

            if isinstance(value, (int, float)):
                formatted_value = f"\033[36m{value_str}\033[0m"
            elif isinstance(value, str):
                formatted_value = f"\033[32m{value_str}\033[0m"
            elif isinstance(value, (list, dict)):
                formatted_value = f"\033[33m{value_str}\033[0m"
            else:
                formatted_value = f"\033[34m{value_str}\033[0m"
        return f"  {formatted_key}: {formatted_value}"

    def format_str(self, model: BaseModel) -> str:
        header = self.format_class_name(model.__class__.__name__) + "\n"

        fields = model.model_dump(mode="json")
        items = [self.format_field(k, v) for k, v in fields.items()]

        return header + "\n".join(items)

    def format_repr(self, model: BaseModel) -> str:
        return self.format_str(model)


class PydanticFormatterMixin:
    __display_formatter__: ClassVar[PydanticFormatter] = ANSIPydanticFormatter()

    def __str__(self) -> str:
        return self.__display_formatter__.format_str(self)

    def __repr__(self) -> str:
        return self.__display_formatter__.format_repr(self)

    def _repr_html_(self) -> str:
        return self.__display_formatter__.format_html(self)

    def _repr_markdown_(self) -> str:
        return self.__display_formatter__.format_markdown(self)



from typing import Any, Generic, Literal, Optional, TypeVar
from uuid import UUID

from pydantic import BaseModel, Field

T = TypeVar("T", bound=BaseModel)


class ItemList(BaseModel, Generic[T]):
    # Used by get_all endpoints
    items: list[T]


class GetAllRequest(BaseModel):
    limit: Optional[int] = None
    offset: int = 0
    filters: dict[str, Any] = Field(default_factory=dict)
    order_by: Optional[str] = "created_at"
    sort_order: Literal["desc", "asc"] = "desc"


class GetOneRequest(BaseModel):
    uid: UUID | None = None
    filters: dict[str, Any] = Field(default_factory=dict)



import random
import string
from typing import List

from syft_rds.utils.resources import load_resource

_adjectives: List[str] | None = None
_nouns: List[str] | None = None


def _load_words(filename: str) -> List[str]:
    content = load_resource(filename)
    return [line.strip() for line in content.splitlines() if line.strip()]


def generate_name() -> str:
    """Generate a Docker-like name using random adjective and noun combinations,
    followed by 4 random alphanumeric characters."""
    global _adjectives, _nouns
    if _adjectives is None:
        _adjectives = _load_words("adjectives.txt")
    if _nouns is None:
        _nouns = _load_words("nouns.txt")

    # Generate 4 random alphanumeric characters
    suffix = "".join(random.choices(string.ascii_lowercase + string.digits, k=4))
    return f"{random.choice(_adjectives)}_{random.choice(_nouns)}_{suffix}"



from uuid import UUID

from syft_rds.client.local_stores.base import CRUDLocalStore
from syft_rds.models import Job, JobCreate, JobUpdate


class JobLocalStore(CRUDLocalStore[Job, JobCreate, JobUpdate]):
    ITEM_TYPE = Job

    def delete_by_id(self, uid: UUID) -> bool:
        """Delete a job by its UUID.

        Args:
            uid: UUID of the job to delete

        Returns:
            True if the job was deleted, False if not found
        """
        return self.store.delete(uid)



from pathlib import Path
from uuid import UUID

from IPython.display import HTML, display
from pydantic import Field
from syft_core import SyftBoxURL

from syft_rds.display_utils.html_format import create_html_repr
from syft_rds.models.base import ItemBase, ItemBaseCreate, ItemBaseUpdate


class Dataset(ItemBase):
    __schema_name__ = "dataset"
    __table_extra_fields__ = [
        "name",
        "summary",
    ]

    name: str = Field(description="Name of the dataset.")
    private: SyftBoxURL = Field(description="Private Syft URL of the dataset.")
    mock: SyftBoxURL = Field(description="Mock Syft URL of the dataset.")
    summary: str | None = Field(description="Summary string of the dataset.")
    readme: SyftBoxURL | None = Field(description="REAMD.md Syft URL of the dataset.")
    tags: list[str] = Field(description="Tags for the dataset.")
    runtime_id: UUID | None = Field(
        default=None, description="ID of the default runtime for the dataset."
    )

    @property
    def mock_path(self) -> Path:
        return self.get_mock_path()

    @property
    def private_path(self) -> Path:
        return self.get_private_path()

    @property
    def readme_path(self) -> Path:
        return self.get_readme_path()

    def get_mock_path(self) -> Path:
        mock_path: Path = self.mock.to_local_path(
            datasites_path=self._syftbox_client.datasites
        )
        if not mock_path.exists():
            raise FileNotFoundError(f"Mock file not found at {mock_path}")
        return mock_path

    def get_private_path(self) -> Path:
        """
        Will always raise PermissionError for non-admin users since they
        don't have permission to access private data.
        """
        # Check if user is admin before attempting to access private path
        if not self._is_admin():
            raise PermissionError(
                f"You must be the datasite admin to access private data. "
                f"Your SyftBox email: '{self._syftbox_client.email}'. "
                f"Host email: '{self._client.config.host}'"
            )

        private_path: Path = self.private.to_local_path(
            datasites_path=self._syftbox_client.datasites
        )
        if not private_path.exists():
            raise FileNotFoundError(
                f"Private data not found at {private_path}. "
                f"Probably you don't have admin permission to the dataset."
            )
        return private_path

    def get_readme_path(self) -> Path:
        """
        Will always raise FileNotFoundError for non-admin since the
        private path will never by synced
        """
        readme_path: Path = self.readme.to_local_path(
            datasites_path=self._syftbox_client.datasites
        )
        if not readme_path.exists():
            raise FileNotFoundError(f"Readme file not found at {readme_path}")
        return readme_path

    def get_description(self) -> str:
        # read the description .md file
        with open(self.get_readme_path()) as f:
            return f.read()

    def describe(self):
        field_to_include = [
            "uid",
            "created_at",
            "updated_at",
            "name",
            "readme_path",
            "mock_path",
        ]
        try:
            # Only include private path if it exists and user has permission
            _ = self.private_path
            field_to_include.append("private_path")
        except (FileNotFoundError, PermissionError):
            pass

        description = create_html_repr(
            obj=self,
            fields=field_to_include,
            display_paths=["mock_path", "readme_path"],
        )

        display(HTML(description))

    def _is_admin(self) -> bool:
        """Check if the current user is admin by comparing email with host."""
        return self._client.email == self._client.host


class DatasetUpdate(ItemBaseUpdate[Dataset]):
    pass


class DatasetCreate(ItemBaseCreate[Dataset]):
    name: str = Field(description="Name of the dataset.")
    path: str = Field(description="Private path of the dataset.")
    mock_path: str = Field(description="Mock path of the dataset.")
    summary: str | None = Field(description="Summary string of the dataset.")
    description_path: str | None = Field(
        description="Path to the detailed REAMD.md of the dataset."
    )
    tags: list[str] | None = Field(description="Tags for the dataset.")
    runtime_id: UUID | None = Field(
        default=None, description="ID of the default runtime for the dataset."
    )



from syft_core import SyftBoxURL
from syft_event import SyftEvents
from syft_event.types import Request

from syft_rds.models import (
    CustomFunction,
    CustomFunctionCreate,
    GetAllRequest,
    GetOneRequest,
    ItemList,
)
from syft_rds.server.router import RPCRouter
from syft_rds.server.services.public_file_service import PublicFileService
from syft_rds.store import YAMLStore
from syft_rds.utils.zip_utils import extract_zip

custom_function_router = RPCRouter()


@custom_function_router.on_request("/create")
def create_custom_function(
    create_request: CustomFunctionCreate, app: SyftEvents, request: Request
) -> CustomFunction:
    custom_function_store: YAMLStore[CustomFunction] = app.state[
        "custom_function_store"
    ]
    public_file_service: PublicFileService = app.state["public_file_service"]
    user = request.sender  # TODO auth

    custom_function = create_request.to_item(extra={"created_by": user})
    custom_function_dir = public_file_service.dir_for_item(item=custom_function)

    if create_request.files_zipped is not None:
        extract_zip(create_request.files_zipped, custom_function_dir)

    custom_function.dir_url = SyftBoxURL.from_path(
        custom_function_dir, app.client.workspace
    )

    return custom_function_store.create(custom_function)


@custom_function_router.on_request("/get_one")
def get_custom_function(request: GetOneRequest, app: SyftEvents) -> CustomFunction:
    custom_function_store: YAMLStore[CustomFunction] = app.state[
        "custom_function_store"
    ]
    filters = request.filters
    if request.uid is not None:
        filters["uid"] = request.uid
    item = custom_function_store.get_one(**filters)
    if item is None:
        raise ValueError(f"No CustomFunction found with filters {filters}")
    return item


@custom_function_router.on_request("/get_all")
def get_all_custom_functions(
    req: GetAllRequest, app: SyftEvents
) -> ItemList[CustomFunction]:
    custom_function_store: YAMLStore[CustomFunction] = app.state[
        "custom_function_store"
    ]
    items = custom_function_store.get_all(
        limit=req.limit,
        offset=req.offset,
        order_by=req.order_by,
        sort_order=req.sort_order,
        filters=req.filters,
    )
    return ItemList[CustomFunction](items=items)


@custom_function_router.on_request("/update")
def update_custom_function(update_request, app: SyftEvents) -> CustomFunction:
    custom_function_store: YAMLStore[CustomFunction] = app.state[
        "custom_function_store"
    ]
    existing_item = custom_function_store.get_by_uid(update_request.uid)
    if existing_item is None:
        raise ValueError(f"CustomFunction with uid {update_request.uid} not found")
    updated_item = existing_item.apply_update(update_request)
    return custom_function_store.update(updated_item.uid, updated_item)



import importlib.resources
from functools import lru_cache

ASSETS = "syft_rds.assets"
CSS_ASSETS = f"{ASSETS}.css"
JS_ASSETS = f"{ASSETS}.js"
SVG_ASSETS = f"{ASSETS}.svg"


def load_css(fname: str) -> str:
    return load_resource(fname, CSS_ASSETS)


def load_js(fname: str) -> str:
    return load_resource(fname, JS_ASSETS)


def load_svg(fname: str) -> str:
    return load_resource(fname, SVG_ASSETS)


@lru_cache(maxsize=64)
def load_resource(fname: str, module: str = ASSETS) -> str:
    return importlib.resources.read_text(module, fname)



from copy import deepcopy

import nh3


def sanitize_html(html_str: str) -> str:
    policy = {
        "tags": ["svg", "strong", "rect", "path", "circle", "code", "pre"],
        "attributes": {
            "*": {"class", "style"},
            "svg": {
                "class",
                "style",
                "xmlns",
                "width",
                "height",
                "viewBox",
                "fill",
                "stroke",
                "stroke-width",
            },
            "path": {"d", "fill", "stroke", "stroke-width"},
            "rect": {"x", "y", "width", "height", "fill", "stroke", "stroke-width"},
            "circle": {"cx", "cy", "r", "fill", "stroke", "stroke-width"},
        },
        "remove": {"script", "style"},
    }

    tags = nh3.ALLOWED_TAGS
    for tag in policy["tags"]:
        tags.add(tag)

    _attributes = deepcopy(nh3.ALLOWED_ATTRIBUTES)
    attributes = {**_attributes, **policy["attributes"]}  # type: ignore

    return nh3.clean(
        html_str,
        tags=tags,
        clean_content_tags=policy["remove"],
        attributes=attributes,
    )



import datetime
from collections import defaultdict
from collections.abc import Iterable
from typing import Any, Dict, List, Mapping, Optional, Union
from uuid import UUID

from loguru import logger

from syft_rds.display_utils.jupyter.icons import Icon
from syft_rds.utils.sanitize import sanitize_html

TABLE_INDEX_KEY = "_table_repr_index"
TABLE_EXTRA_FIELDS = "__table_extra_fields__"
TABLE_COL_WIDTHS = "__table_col_widths__"
CUSTOM_ROW_REPR = "__table_row_repr__"
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"

ID_FIELD = "uid"
DATE_FIELD = "created_at"
TYPE_FIELD = "type"
MAPPING_KEY_FIELD = "key"
RESERVED_COLUMNS = {
    ID_FIELD,
    TYPE_FIELD,
    MAPPING_KEY_FIELD,
    TABLE_INDEX_KEY,
    DATE_FIELD,
}


def _get_values_for_table_repr(obj: Any) -> list:
    if isinstance(obj, Mapping):
        values = list(obj.values())
    elif isinstance(obj, list):
        values = obj

    return values


def _make_template_columns(first_value: Any) -> tuple[str | None, str | None]:
    grid_template_cols = getattr(first_value, TABLE_COL_WIDTHS, None)
    if isinstance(grid_template_cols, list):
        grid_template_columns = " ".join(grid_template_cols)
        grid_template_cell_columns = "unset"
    else:
        grid_template_columns = None
        grid_template_cell_columns = None
    return grid_template_columns, grid_template_cell_columns


def _get_extra_field_value(obj: Any, field: str) -> Any:
    """
    Retrieves the value of an extra field from an object.
    Extra fields can be nested attributes separated by '.'.

    Args:
        obj (Any): The object from which to retrieve the extra field.
        field (str): The name of the extra field to retrieve.

    Returns:
        Any: The value of the extra field.
    """
    attrs = field.split(".")
    value = obj
    for attr in attrs:
        value = getattr(value, attr, None)
        if value is None:
            break
    return value


def _create_table_rows(
    items: Union[Mapping, Iterable],
    is_homogenous: bool,
    extra_fields: Optional[List[str]] = None,
    add_index: bool = True,
) -> List[Dict[str, Any]]:
    """
    Creates row data for a table based on input object obj.

    If valid table data cannot be created, an empty list is returned.

    Args:
        items (Union[Mapping, Iterable]): The input data as a Mapping or Iterable.
        is_homogenous (bool): A boolean indicating whether the data is homogenous.
        extra_fields (Optional[List], optional): Additional fields to include in the table. Defaults to None.
        add_index (bool, optional): Whether to add an index column. Defaults to True.

    Returns:
        List[Dict[str, Any]]: A list of dictionaries where each dictionary represents a row in the table.
    """
    extra_fields = extra_fields or []
    column_data = defaultdict(list)

    for item in iter(items.items() if isinstance(items, Mapping) else items):
        # Handle mapping items by extracting key and value
        if isinstance(items, Mapping):
            key, item = item
            column_data[MAPPING_KEY_FIELD].append(key)

        # Add ID field if present
        uid = getattr(item, ID_FIELD, None)
        if uid is not None:
            column_data[ID_FIELD].append(uid)

        # Add date_created field if present
        if hasattr(item, DATE_FIELD):
            column_data[DATE_FIELD].append(getattr(item, DATE_FIELD))

        # Add type information for heterogeneous collections
        if not is_homogenous:
            column_data["type"].append(item.__class__.__name__)

        # Process custom row representation if available
        if hasattr(item, CUSTOM_ROW_REPR):
            row_dict = getattr(item, CUSTOM_ROW_REPR)()
            for col_name, value in row_dict.keys():
                if col_name not in RESERVED_COLUMNS:
                    column_data[col_name].append(value)
        else:
            # Process extra fields
            for col_name in extra_fields:
                if col_name in RESERVED_COLUMNS:
                    continue

                try:
                    value = _get_extra_field_value(item, col_name)
                except Exception:
                    value = None

                column_data[col_name].append(value)

    # Validate that all columns have the same length
    unique_col_lengths = {len(column_data[col]) for col in column_data.keys()}
    if not column_data or len(unique_col_lengths) != 1:
        logger.debug("Cannot create table for items with different number of fields.")
        return []

    # Add index column if requested
    num_rows = unique_col_lengths.pop()
    if add_index and TABLE_INDEX_KEY not in column_data:
        column_data[TABLE_INDEX_KEY] = list(range(num_rows))

    # Transpose column data to row data
    row_data = []
    for i in range(num_rows):
        row = {}
        for col_name in column_data:
            row[col_name] = column_data[col_name][i]
        row_data.append(row)

    return row_data


def prepare_table_data(
    obj: Any,
    add_index: bool = True,
) -> tuple[list[dict], dict]:
    """
    Creates table data and metadata for a given object.

    If a tabular representation cannot be created, an empty list and empty dict are returned instead.

    Args:
        obj (Any): The input object for which table data is prepared.
        add_index (bool, optional): Whether to add an index column to the table. Defaults to True.

    Returns:
        tuple: A tuple (table_data, table_metadata) where table_data is a list of dictionaries
        where each dictionary represents a row in the table and table_metadata is a dictionary
        containing metadata about the table such as name, icon, etc.

    """

    values = _get_values_for_table_repr(obj)
    if len(values) == 0:
        return [], {}

    # check first value and obj itself to see if syft in mro. If not, don't create table
    first_value = values[0]

    extra_fields = getattr(first_value, TABLE_EXTRA_FIELDS, [])
    is_homogenous = len({type(x) for x in values}) == 1
    if is_homogenous:
        cls_name = first_value.__class__.__name__
        grid_template_columns, grid_template_cell_columns = _make_template_columns(
            first_value
        )
    else:
        cls_name = ""
        grid_template_columns = None
        grid_template_cell_columns = None

    table_data = _create_table_rows(
        items=obj,
        is_homogenous=is_homogenous,
        extra_fields=extra_fields,
        add_index=add_index,
    )
    # if empty result, collection objects have no table representation
    if not table_data:
        return [], {}

    table_metadata = {
        "name": f"{cls_name} {obj.__class__.__name__.capitalize()}",
        "columns": list(table_data[0].keys()),
        "icon": Icon.TABLE.svg,
        "grid_template_columns": grid_template_columns,
        "grid_template_cell_columns": grid_template_cell_columns,
    }

    return table_data, table_metadata


def format_dict(data: Any) -> str:
    if not isinstance(data, dict):
        return data

    # is_component_dict = set(data.keys()) == {"type", "value"}
    # if is_component_dict and "badge" in data["type"]:
    #     return Badge(value=data["value"], badge_class=data["type"]).to_html()
    # elif is_component_dict and "label" in data["type"]:
    #     return Label(value=data["value"], label_class=data["type"]).to_html()
    # if is_component_dict and "clipboard" in data["type"]:
    #     return CopyButton(copy_text=data["value"]).to_html()

    return sanitize_html(str(data))


def format_uid(uid: UUID) -> str:
    return str(uid)
    # return CopyButton(copy_text=str(uid)).to_html()


def format_table_value(value: Any) -> str:
    """
    Format a single cell value for display in a table.
    TODO add support for more complex types like components from PySyft.
    """
    if isinstance(value, UUID):
        return format_uid(value)
    elif isinstance(value, datetime.datetime):
        return value.strftime(DATETIME_FORMAT)
    elif isinstance(value, dict):
        return format_dict(value)
    else:
        return sanitize_html(str(value).replace("\n", "<br>"))


def format_table_data(table_data: list[dict[str, Any]]) -> list[dict[str, str]]:
    formatted: list[dict[str, str]] = []
    for row in table_data:
        row_formatted: dict[str, str] = {}
        for k, v in row.items():
            row_formatted[k] = format_table_value(v)
        formatted.append(row_formatted)
    return formatted



from loguru import logger
import os
from pathlib import Path

from syft_rds.client.rds_clients.base import RDSClientModule
from syft_rds.models import (
    Runtime,
    RuntimeCreate,
    RuntimeKind,
    RuntimeConfig,
    PythonRuntimeConfig,
    DockerRuntimeConfig,
    KubernetesRuntimeConfig,
)

DEFAULT_RUNTIME_KIND = os.getenv("SYFT_RDS_DEFAULT_RUNTIME_KIND", "python")
DEFAULT_RUNTIME_NAME = os.getenv(
    "SYFT_RDS_DEFAULT_RUNTIME_NAME", "syft_default_python_runtime"
)
PROJECT_ROOT = Path(__file__).resolve().parents[4]
DEFAULT_DOCKERFILE_FILE_PATH = PROJECT_ROOT / "runtimes" / "python.Dockerfile"


class RuntimeRDSClient(RDSClientModule[Runtime]):
    ITEM_TYPE = Runtime

    def create(
        self,
        runtime_name: str | None = None,
        runtime_kind: str | None = None,
        config: dict | None = None,
        description: str | None = None,
    ) -> Runtime:
        """
        Create a runtime.
        """
        if runtime_name is None and runtime_kind is None:
            return self._get_or_create_default()

        self._validate_runtime_args(runtime_name, runtime_kind, config)

        # TODO: Uncomment this
        # if self.get_runtime_by_name(runtime_name):
        #     raise ValueError(f"Runtime with name '{runtime_name}' already exists.")

        runtime_config: RuntimeConfig = self._create_runtime_config(
            runtime_kind, config
        )
        runtime_create: RuntimeCreate = RuntimeCreate(
            name=runtime_name,
            kind=RuntimeKind(runtime_kind),
            config=runtime_config,
            description=description,
        )

        return self.get_or_create(runtime_create)

    def get_runtime_by_name(self, name: str) -> Runtime | None:
        try:
            return self.get(name=name)
        except Exception as e:
            logger.debug(f"Error getting runtime by name: {e}")
            return None

    def get_or_create(self, runtime_create: RuntimeCreate) -> Runtime:
        fetched_runtime = self.get_runtime_by_name(runtime_create.name)

        if fetched_runtime:
            logger.debug(
                f"Runtime '{fetched_runtime.name}' already exists. Returning existing runtime."
            )
            return fetched_runtime

        runtime = self.rpc.runtime.create(runtime_create)
        logger.debug(f"Runtime created: {runtime}")

        return runtime

    def _create_runtime_config(
        self, runtime_kind: str, config: dict | None = None
    ) -> RuntimeConfig:
        if config is None:
            config = {}

        kind_str = runtime_kind.lower()

        config_map = {
            RuntimeKind.PYTHON.value: PythonRuntimeConfig,
            RuntimeKind.DOCKER.value: DockerRuntimeConfig,
            RuntimeKind.KUBERNETES.value: KubernetesRuntimeConfig,
        }

        config_class = config_map.get(kind_str)

        if config_class:
            return config_class(**config)
        else:
            raise ValueError(f"Unsupported runtime type: {runtime_kind}")

    def _get_or_create_default(self) -> Runtime:
        """
        Get the default runtime if it exists, otherwise create it.
        The default runtime is a python runtime with the python 3.12.
        """
        try:
            default_runtime_create = RuntimeCreate(
                kind=DEFAULT_RUNTIME_KIND,
                name=DEFAULT_RUNTIME_NAME,
                config=PythonRuntimeConfig(
                    version="3.12",
                ),
                description="Default Python runtime for Syft RDS",
            )
            return self.get_or_create(default_runtime_create)
        except Exception as e:
            logger.error(f"Error getting or creating default runtime: {e}")
            raise e

    def _validate_runtime_args(self, runtime_name, runtime_kind, config):
        if runtime_name is not None and runtime_kind is None:
            raise ValueError(
                "Runtime kind must be provided if runtime name is provided"
            )
        if runtime_kind is not None and runtime_kind not in [
            r.value for r in RuntimeKind
        ]:
            raise ValueError(
                f"Invalid runtime kind: {runtime_kind}. Must be one of {RuntimeKind}"
            )
        if config is not None and runtime_kind is None:
            raise ValueError("Runtime kind must be provided if config is provided")



import base64
import enum
from pathlib import Path
from typing import Optional, TypeVar

from IPython.display import HTML, display
from pydantic import (
    field_serializer,
    field_validator,
)
from syft_core import SyftBoxURL

from syft_rds.display_utils.html_format import create_html_repr
from syft_rds.models.base import ItemBase, ItemBaseCreate, ItemBaseUpdate

T = TypeVar("T", bound=ItemBase)

MAX_USERCODE_ZIP_SIZE = 1  # MB


class UserCodeType(enum.Enum):
    FILE = "file"
    FOLDER = "folder"


class UserCode(ItemBase):
    __schema_name__ = "usercode"
    __table_extra_fields__ = [
        "name",
    ]

    name: str
    dir_url: SyftBoxURL | None = None
    code_type: UserCodeType
    entrypoint: str

    @property
    def local_dir(self) -> Path:
        if self.dir_url is None:
            raise ValueError("dir_url is not set")
        client = self._client
        return self.dir_url.to_local_path(
            datasites_path=client._syftbox_client.datasites
        )

    def describe(self) -> None:
        html_description = create_html_repr(
            obj=self,
            fields=[
                "uid",
                "created_by",
                "created_at",
                "updated_at",
                "name",
                "local_dir",
                "code_type",
                "entrypoint",
            ],
            display_paths=["local_dir"],
        )
        display(HTML(html_description))


class UserCodeCreate(ItemBaseCreate[UserCode]):
    name: Optional[str] = None
    files_zipped: bytes | None = None
    code_type: UserCodeType
    entrypoint: str

    @field_serializer("files_zipped")
    def serialize_to_str(self, v: bytes | None) -> str | None:
        # Custom serialize for zipped binary data
        if v is None:
            return None
        return base64.b64encode(v).decode()

    @field_validator("files_zipped", mode="before")
    @classmethod
    def deserialize_from_str(cls, v):
        # Custom deserialize for zipped binary data
        if isinstance(v, str):
            return base64.b64decode(v)
        return v

    @field_validator("files_zipped", mode="after")
    @classmethod
    def validate_code_size(cls, v: bytes) -> bytes:
        zip_size_mb = len(v) / 1024 / 1024
        if zip_size_mb > MAX_USERCODE_ZIP_SIZE:
            raise ValueError(
                f"Provided files too large: {zip_size_mb:.2f}MB. Max size is {MAX_USERCODE_ZIP_SIZE}MB"
            )
        return v


class UserCodeUpdate(ItemBaseUpdate[UserCode]):
    pass



import enum
import json
import subprocess
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional
from uuid import UUID

from IPython.display import HTML, display
from loguru import logger
from pydantic import BaseModel, Field, model_validator
from syft_core import SyftBoxURL

from syft_rds.display_utils.html_format import create_html_repr
from syft_rds.models.base import ItemBase, ItemBaseCreate, ItemBaseUpdate
from syft_rds.models.runtime_models import Runtime
from syft_rds.utils.name_generator import generate_name

if TYPE_CHECKING:
    from syft_rds.models import CustomFunction, UserCode


class JobStatus(str, enum.Enum):
    pending_code_review = "pending_code_review"
    job_run_failed = "job_run_failed"
    job_run_finished = "job_run_finished"
    job_in_progress = "job_in_progress"

    # end states
    rejected = "rejected"  # failed to pass the review
    shared = "shared"  # shared with the user


class JobErrorKind(str, enum.Enum):
    no_error = "no_error"
    timeout = "timeout"
    cancelled = "cancelled"
    execution_failed = "execution_failed"
    failed_code_review = "failed_code_review"
    failed_output_review = "failed_output_review"


class Job(ItemBase):
    class Config:
        extra = "forbid"

    __schema_name__ = "job"
    __table_extra_fields__ = [
        "created_by",
        "name",
        "dataset_name",
        "runtime_name",
        "status",
        "error",
    ]

    name: str = Field(default_factory=generate_name)
    dataset_name: str
    runtime_id: UUID
    user_code_id: UUID
    custom_function_id: Optional[UUID] = None
    description: str | None = None
    tags: list[str] = Field(default_factory=list)
    user_metadata: dict = {}
    status: JobStatus = JobStatus.pending_code_review
    error: JobErrorKind = JobErrorKind.no_error
    error_message: str | None = None
    output_url: SyftBoxURL | None = None

    def describe(self) -> None:
        html_description = create_html_repr(
            obj=self,
            fields=[
                "uid",
                "created_by",
                "created_at",
                "updated_at",
                "name",
                "description",
                "status",
                "error",
                "error_message",
                "output_path",
                "dataset_name",
                "user_code_name",
                "custom_function_name",
            ],
            display_paths=["output_path"],
        )
        display(HTML(html_description))

    @property
    def runtime(self) -> "Runtime":
        """Get the runtime of the job"""
        client = self._client
        return client.runtime.get(self.runtime_id)

    @property
    def runtime_name(self) -> str:
        """Get the name of the runtime of the job"""
        return self.runtime.name

    @property
    def user_code(self) -> "UserCode":
        client = self._client
        return client.user_code.get(self.user_code_id)

    @property
    def user_code_name(self) -> str:
        return self.user_code.name

    @property
    def custom_function(self) -> "Optional[CustomFunction]":
        if self.custom_function_id is None:
            return None
        client = self._client
        return client.custom_function.get(self.custom_function_id)

    @property
    def custom_function_name(self) -> Optional[str]:
        if self.custom_function is None:
            return None
        return self.custom_function.name

    def show_user_code(self) -> None:
        user_code = self.user_code
        user_code.describe()

    def get_update_for_in_progress(self) -> "JobUpdate":
        return JobUpdate(
            uid=self.uid,
            status=JobStatus.job_in_progress,
        )

    def get_update_for_return_code(
        self, return_code: int | subprocess.Popen, error_message: str | None = None
    ) -> "JobUpdate":
        if not isinstance(return_code, int):
            return self.get_update_for_in_progress()
        if return_code == 0:
            self.status = JobStatus.job_run_finished
            self.error = JobErrorKind.no_error
            self.error_message = None
        else:
            self.status = JobStatus.job_run_failed
            self.error = JobErrorKind.execution_failed
            self.error_message = error_message

        return JobUpdate(
            uid=self.uid,
            status=self.status,
            error=self.error,
            error_message=self.error_message,
        )

    @property
    def output_path(self) -> Path:
        return self.get_output_path()

    def get_output_path(self) -> Path:
        if self.output_url is None:
            raise ValueError("output_url is not set")
        client = self._client
        return self.output_url.to_local_path(
            datasites_path=client._syftbox_client.datasites
        )

    @model_validator(mode="after")
    def validate_status(self):
        if (
            self.status == JobStatus.job_run_failed
            and self.error == JobErrorKind.no_error
        ):
            raise ValueError("error must be set if status is failed")
        return self


class JobUpdate(ItemBaseUpdate[Job]):
    status: Optional[JobStatus] = None
    error: Optional[JobErrorKind] = None
    error_message: Optional[str] = None


class JobCreate(ItemBaseCreate[Job]):
    dataset_name: str
    user_code_id: UUID
    runtime_id: UUID
    name: str | None = None
    description: str | None = None
    tags: list[str] = Field(default_factory=list)
    custom_function_id: Optional[UUID] = None


class JobConfig(BaseModel):
    """Configuration for a job run"""

    function_folder: Path
    args: list[str]
    data_path: Path
    runtime: "Runtime"
    job_folder: Optional[Path] = Field(
        default_factory=lambda: Path("jobs") / datetime.now().strftime("%Y%m%d_%H%M%S")
    )
    timeout: int = 60
    data_mount_dir: str = "/app/data"
    extra_env: dict[str, str] = {}
    blocking: bool = Field(default=True)

    @property
    def job_path(self) -> Path:
        """Derived path for job folder"""
        return Path(self.job_folder)

    @property
    def logs_dir(self) -> Path:
        """Derived path for logs directory"""
        return self.job_path / "logs"

    @property
    def output_dir(self) -> Path:
        """Derived path for output directory"""
        return self.job_path / "output"

    def get_env(self) -> dict[str, str]:
        return self.extra_env | self._base_env

    def get_env_as_docker_args(self) -> list[str]:
        return [f"-e {k}={v}" for k, v in self.get_env().items()]

    def get_extra_env_as_docker_args(self) -> list[str]:
        return [f"-e {k}={v}" for k, v in self.extra_env.items()]

    @property
    def _base_env(self) -> dict[str, str]:
        interpreter = " ".join(self.runtime.cmd)
        # interpreter_str = f"'{interpreter}'" if " " in interpreter else interpreter
        return {
            "OUTPUT_DIR": str(self.output_dir.absolute()),
            "DATA_DIR": str(self.data_path.absolute()),
            "CODE_DIR": str(self.function_folder.absolute()),
            "TIMEOUT": str(self.timeout),
            "INPUT_FILE": str(self.function_folder / self.args[0]),
            "INTERPRETER": interpreter,
        }


class JobResults(BaseModel):
    _MAX_LOADED_FILE_SIZE: int = 10 * 1024 * 1024  # 10 MB

    job: Job
    results_dir: Path

    @property
    def logs_dir(self) -> Path:
        return self.results_dir / "logs"

    @property
    def output_dir(self) -> Path:
        return self.results_dir / "output"

    @property
    def stderr_file(self) -> Path:
        return self.logs_dir / "stderr.log"

    @property
    def stdout_file(self) -> Path:
        return self.logs_dir / "stdout.log"

    @property
    def stderr(self) -> str | None:
        if self.stderr_file.exists():
            return self.stderr_file.read_text()
        return None

    @property
    def stdout(self) -> str | None:
        if self.stdout_file.exists():
            return self.stdout_file.read_text()
        return None

    @property
    def log_files(self) -> list[Path]:
        return list(self.logs_dir.glob("*"))

    @property
    def output_files(self) -> list[Path]:
        return list(self.output_dir.glob("*"))

    @property
    def outputs(self) -> dict[str, Any]:
        outputs = {}
        for file in self.output_dir.glob("*"):
            try:
                contents = load_output_file(
                    filepath=file, max_size=self._MAX_LOADED_FILE_SIZE
                )
                outputs[file.name] = contents
            except ValueError as e:
                logger.warning(
                    f"Skipping output {file.name}: {e}. Please load this file manually."
                )
                continue
        return outputs

    def describe(self):
        display_paths = ["output_dir"]
        if self.stdout_file.exists():
            display_paths.append("stdout_file")
        if self.stderr_file.exists():
            display_paths.append("stderr_file")

        html_repr = create_html_repr(
            obj=self,
            fields=["output_dir", "logs_dir"],
            display_paths=display_paths,
        )

        display(HTML(html_repr))


def load_output_file(filepath: Path, max_size: int) -> Any:
    if not filepath.exists():
        raise ValueError(f"File {filepath} does not exist.")

    file_size = filepath.stat().st_size
    if file_size > max_size:
        raise ValueError(
            f"File the maximum size of {int(max_size / (1024 * 1024))} MB."
        )

    if filepath.suffix == ".json":
        with open(filepath, "r") as f:
            return json.load(f)

    elif filepath.suffix == ".parquet":
        import pandas as pd

        return pd.read_parquet(filepath)

    elif filepath.suffix == ".csv":
        import pandas as pd

        return pd.read_csv(filepath)

    elif filepath.suffix in {".txt", ".log", ".md", ".html"}:
        with open(filepath, "r") as f:
            return f.read()

    else:
        raise ValueError("Unsupported file type.")



import enum
import html
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional

import jinja2
from pydantic import BaseModel
from rich.console import Console
from rich.tree import Tree

from syft_rds.utils.resources import load_css

jinja_env = jinja2.Environment(loader=jinja2.PackageLoader("syft_rds", "assets/jinja"))  # nosec
PERM_FILE = "syftperm.yaml"


def make_dirtree_string(root_dir: Path) -> Optional[str]:
    try:
        # Create a Tree object
        tree = Tree(f"📁 {root_dir.name}")

        def add_dir(tree: Tree, path: Path) -> None:
            for child in sorted(path.iterdir(), key=lambda p: (not p.is_dir(), p.name)):
                if child.is_dir():
                    sub_tree = tree.add(f"📁 {child.name}")
                    add_dir(sub_tree, child)
                elif child.name == PERM_FILE:
                    tree.add(f"🛡️ {child.name}")
                else:
                    tree.add(f"📄 {child.name}")

        add_dir(tree, root_dir)
        console = Console()
        with console.capture() as capture:
            console.print(tree)
        return capture.get()
    except Exception as e:
        return f"Could not generate directory tree: {e}"


def format_field_value(value: Any) -> str:
    """Format various field types for display with multiline text for paths"""
    if value is None:
        return "None"
    elif isinstance(value, datetime):
        return value.strftime("%Y-%m-%d %H:%M:%S")
    elif isinstance(value, Path):
        path_str = str(value)
        return f'<div class="path-text">{html.escape(path_str)}</div>'
    elif isinstance(value, enum.Enum):
        return str(value.name)
    elif hasattr(value, "uid"):
        return f"{type(value).__name__}({value.uid})"
    else:
        return f'<div class="path-text">{html.escape(str(value))}</div>'


def prepare_path_display(obj: Any, path_field: str, open_default: bool = False) -> str:
    """
    Generate HTML for displaying a path (either file content or directory tree)

    Args:
        obj: The object containing the path field
        path_field: The name of the field containing the path
        open_default: Whether the details element should be open by default

    Returns:
        str: HTML representation of the file content or directory tree
    """
    template = jinja_env.get_template("model_repr_file_section.jinja2")
    render_params = {"title": path_field, "open": open_default}

    # Check if field exists
    if not hasattr(obj, path_field):
        render_params["error"] = f"Field '{path_field}' does not exist on this object."
        render_params["error_class"] = "error-text"
        return template.render(**render_params)

    # Get path
    path = getattr(obj, path_field)
    if not isinstance(path, Path) or not path.exists():
        render_params["error"] = (
            "The path does not exist or is not a valid Path object."
        )
        render_params["error_class"] = "warning-text"
        return template.render(**render_params)

    try:
        if path.is_file():
            # Handle file display
            with open(path, "r") as f:
                file_content = f.read()

            render_params["title"] = f"{path_field} content ({path.name})"
            render_params["content"] = html.escape(file_content)
        elif path.is_dir():
            # Handle directory tree display
            tree_content = make_dirtree_string(path)
            render_params["title"] = f"{path_field} directory structure ({path.name})"
            render_params["content"] = html.escape(tree_content) if tree_content else ""
        else:
            render_params["error"] = "Path is neither a file nor a directory."
            render_params["error_class"] = "warning-text"
    except Exception as e:
        render_params["error"] = str(e)
        render_params["error_class"] = "error-text"

    return template.render(**render_params)


def create_html_repr(
    obj: BaseModel,
    fields: List[str],
    name_field: str = "name",
    display_paths: Optional[List[str]] = None,
    paths_open: bool = False,
) -> str:
    """
    Generate an HTML representation of a model with automatic path type detection

    Args:
        obj: A Pydantic BaseModel
        fields: List of fields to display
        display_paths: List of field names containing paths that should be displayed
            (automatically detects if it's a file or directory)
        paths_open: Whether path sections should be open by default

    Returns:
        str: HTML representation
    """
    display_paths = display_paths or []

    # Prepare field values
    field_values = {}
    for field in fields:
        if hasattr(obj, field):
            field_values[field] = format_field_value(getattr(obj, field))
        else:
            field_values[field] = '<span class="error-text">Field not found</span>'

    # Get object name
    obj_name = getattr(obj, name_field, None)

    # Prepare path divs (file content and directory trees)
    path_html_reprs = [
        prepare_path_display(obj, path_field, paths_open)
        for path_field in display_paths
    ]

    template = jinja_env.get_template("model_repr.jinja2")
    return template.render(
        obj=obj,
        obj_name=obj_name,
        display_fields=fields,
        field_values=field_values,
        path_displays=path_html_reprs,
        css=load_css("model_repr.css"),
    )



from typing import TYPE_CHECKING
from uuid import UUID

if TYPE_CHECKING:
    from syft_rds.client.rds_client import RDSClient


class GlobalClientRegistry:
    registry = {}

    @classmethod
    def get_client(cls, uid: UUID) -> "RDSClient":
        return cls.registry[uid]

    @classmethod
    def register_client(cls, client: "RDSClient") -> None:
        cls.registry[client.uid] = client



from typing import Final, Type
from uuid import UUID

from syft_rds.client.local_stores.base import CRUDLocalStore
from syft_rds.models import UserCode, UserCodeCreate, UserCodeUpdate


class UserCodeLocalStore(CRUDLocalStore[UserCode, UserCodeCreate, UserCodeUpdate]):
    ITEM_TYPE: Final[Type[UserCode]] = UserCode

    def delete_by_id(self, uid: UUID) -> bool:
        """Delete a user code by its UUID.

        Args:
            uid: UUID of the user code to delete

        Returns:
            True if the user code was deleted, False if not found
        """
        return self.store.delete(uid)



from typing import Dict, List
from syft_rds.models import DockerMount, JobConfig


class MountProvider:
    """Base class for mount providers"""

    def get_mounts(self, job_config: JobConfig) -> List[DockerMount]:
        """Get additional mounts for a job"""
        return []


# Registry to store mount providers
_mount_providers: Dict[str, MountProvider] = {}


def register_mount_provider(app_name: str, provider: MountProvider) -> None:
    """Register a mount provider for an app"""
    _mount_providers[app_name] = provider


def get_mount_provider(app_name: str) -> MountProvider | None:
    """Get a mount provider for an app"""
    return _mount_providers.get(app_name)



from pathlib import Path

from syft_rds.client.rds_clients.base import RDSClientModule
from syft_rds.client.utils import PathLike
from syft_rds.models import (
    UserCode,
    UserCodeCreate,
    UserCodeType,
)
from syft_rds.utils.zip_utils import zip_to_bytes


class UserCodeRDSClient(RDSClientModule[UserCode]):
    ITEM_TYPE = UserCode

    def create(
        self,
        code_path: PathLike,
        name: str | None = None,
        entrypoint: str | None = None,
    ) -> UserCode:
        code_path = Path(code_path)
        if not code_path.exists():
            raise FileNotFoundError(f"Path {code_path} does not exist.")

        if code_path.is_dir():
            code_type = UserCodeType.FOLDER

            # Entrypoint is required for folder-type code
            if not entrypoint:
                raise ValueError("Entrypoint should be provided for folder code.")

            # Validate that the entrypoint exists within the folder
            if not (code_path / entrypoint).exists():
                raise FileNotFoundError(
                    f"Entrypoint {entrypoint} does not exist in {code_path}."
                )
            files_zipped = zip_to_bytes(files_or_dirs=[code_path], base_dir=code_path)
        else:
            code_type = UserCodeType.FILE

            # For file-type code, the entrypoint is the file name
            entrypoint = entrypoint or code_path.name

            files_zipped = zip_to_bytes(files_or_dirs=code_path)

        user_code_create = UserCodeCreate(
            name=name,
            files_zipped=files_zipped,
            code_type=code_type,
            entrypoint=entrypoint,
        )

        user_code = self.rpc.user_code.create(user_code_create)

        return user_code



from io import BytesIO
from pathlib import Path
from typing import Dict, List, Optional, Union
from zipfile import ZipFile

PathLike = Union[str, Path]


def extract_zip(zip_data: bytes, target_dir: PathLike) -> None:
    """Extract zip data to a target directory.

    Args:
        zip_data: Bytes containing zip content
        target_dir: Directory to extract files to
    """
    with ZipFile(BytesIO(zip_data)) as z:
        z.extractall(str(target_dir))


def zip_to_bytes(
    files_or_dirs: Union[PathLike, List[PathLike]], base_dir: Optional[PathLike] = None
) -> bytes:
    """Create a zip file from files or directories, returning the zip content as bytes.

    Args:
        files_or_dirs: Single path or list of paths to include
        base_dir: Optional base directory for relative paths in the zip

    Returns:
        Bytes containing the zip file
    """
    buffer = BytesIO()

    with ZipFile(buffer, "w") as z:
        paths = (
            [files_or_dirs] if isinstance(files_or_dirs, (str, Path)) else files_or_dirs
        )

        for path in paths:
            path = Path(path)

            if not path.exists():
                raise FileNotFoundError(f"Path {path} does not exist.")
            if path.is_file():
                arcname = path.name if base_dir is None else path.relative_to(base_dir)
                z.write(path, arcname=str(arcname))
            elif path.is_dir():
                for file_path in path.rglob("*"):
                    if file_path.is_file():
                        arcname = (
                            file_path.name
                            if base_dir is None
                            else file_path.relative_to(base_dir)
                        )
                        z.write(file_path, arcname=str(arcname))

    return buffer.getvalue()


def get_files_from_zip(zip_data: bytes) -> Dict[str, bytes]:
    """Extract files from zip data to a dictionary.

    Args:
        zip_data: Bytes containing zip content

    Returns:
        Dictionary mapping filenames to file contents
    """
    result = {}
    with ZipFile(BytesIO(zip_data)) as z:
        for filename in z.namelist():
            result[filename] = z.read(filename)

    return result



from enum import Enum
from pathlib import Path
from typing import Literal
import json
import hashlib
from typing import Any

from loguru import logger
from pydantic import (
    BaseModel,
    Field,
    field_validator,
    model_validator,
)

from syft_rds.client.utils import PathLike
from syft_rds.models.base import ItemBase, ItemBaseCreate, ItemBaseUpdate


class RuntimeKind(str, Enum):
    PYTHON = "python"
    DOCKER = "docker"
    KUBERNETES = "kubernetes"


class BaseRuntimeConfig(BaseModel):
    """Base configuration for runtime environments."""

    cmd: list[str] | None = None

    def validate_config(self) -> bool:
        """Override in subclasses for custom validation."""
        return True


class PythonRuntimeConfig(BaseRuntimeConfig):
    version: str | None = None
    requirements_file: PathLike | None = None
    cmd: list[str] = Field(default_factory=lambda: ["python"])

    @field_validator("requirements_file")
    def validate_requirements_file_exist(cls, value: PathLike) -> PathLike:
        if value is not None:
            requirements_file_path = Path(value).expanduser().resolve()
            if not requirements_file_path.exists():
                raise FileNotFoundError(f"Requirements file '{value}' does not exist")
        return value


class DockerMount(BaseModel):
    source: PathLike
    target: PathLike
    mode: Literal["ro", "rw"] = "ro"

    @field_validator("source")
    def validate_source_path_exists(cls, value: PathLike) -> PathLike:
        source_path = Path(value).expanduser().resolve()
        if not source_path.exists():
            # raise FileNotFoundError(f"Source path '{value}' does not exist")
            logger.warning(f"Source path '{value}' does not exist")
        return value


class DockerRuntimeConfig(BaseRuntimeConfig):
    dockerfile: PathLike | None = Field(default=None, exclude=True)
    entrypoint_script: PathLike | None = Field(default=None, exclude=True)
    dockerfile_content: str
    image_name: str | None = None
    cmd: list[str] = ["python"]
    app_name: str | None = None
    extra_mounts: list[DockerMount] = Field(default_factory=list)

    @model_validator(mode="before")
    @classmethod
    def load_content_from_path(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        If a 'dockerfile' path is provided, read its content into 'dockerfile_content'.
        This validator runs before any other field validation.
        """
        if "dockerfile" in values and values["dockerfile"] is not None:
            dockerfile_path = Path(values["dockerfile"]).expanduser().resolve()
            if not dockerfile_path.exists():
                raise FileNotFoundError(
                    f"The specified Dockerfile does not exist at: {dockerfile_path}"
                )
            values["dockerfile_content"] = dockerfile_path.read_text()
        elif "dockerfile_content" not in values:
            raise ValueError(
                "You must provide a path to a Dockerfile via the 'dockerfile' argument."
            )
        return values

    @field_validator("dockerfile_content")
    def validate_dockerfile_content(cls, dockerfile_content: str) -> str:
        if not dockerfile_content:
            raise ValueError("Dockerfile cannot be empty")
        return dockerfile_content.strip()

    def __hash__(self) -> int:
        return hash(self.dockerfile_content)

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, DockerRuntimeConfig):
            return False
        return self.dockerfile_content == __value.dockerfile_content


class KubernetesRuntimeConfig(BaseRuntimeConfig):
    image: str
    namespace: str = "syft-rds"
    num_workers: int = 1
    cmd: list[str] = Field(default_factory=list)


RuntimeConfig = PythonRuntimeConfig | DockerRuntimeConfig | KubernetesRuntimeConfig


class Runtime(ItemBase):
    __schema_name__ = "runtime"
    __table_extra_fields__ = [
        "name",
        "kind",
    ]

    name: str | None = None
    kind: RuntimeKind
    config: RuntimeConfig = Field(default_factory=dict)
    tags: list[str] = Field(default_factory=list)
    description: str | None = Field(default=None)

    @model_validator(mode="after")
    def set_name_with_prefix(self):
        if self.name is not None:
            return self
        # Create a unique name for the runtime based on the config's hash
        # TODO: create hash based on Docker file's content
        config_dict = self.config.model_dump(mode="json")
        config_str = json.dumps(config_dict, sort_keys=True)
        config_hash = hashlib.sha256(config_str.encode()).hexdigest()[:6]
        self.name = f"{self.kind.value.lower()}_{config_hash}"
        return self

    @property
    def cmd(self) -> list[str]:
        return self.config.cmd


class RuntimeCreate(ItemBaseCreate[Runtime]):
    name: str | None = None
    kind: RuntimeKind
    config: RuntimeConfig = Field(default_factory=dict)
    tags: list[str] = Field(default_factory=list)
    description: str | None = Field(default=None)

    @model_validator(mode="after")
    def set_name_with_prefix(self):
        if self.name is not None:
            return self
        # Create a unique name for the runtime based on the config's hash
        config_dict = self.config.model_dump(mode="json")
        config_str = json.dumps(config_dict, sort_keys=True)
        config_hash = hashlib.sha256(config_str.encode()).hexdigest()[:6]
        self.name = f"{self.kind.value.lower()}_{config_hash}"
        return self


class RuntimeUpdate(ItemBaseUpdate[Runtime]):
    pass



__version__ = "0.3.0"

from syft_rds.utils.paths import RDS_NOTEBOOKS_PATH, RDS_REPO_PATH  # noqa
from syft_rds.display_utils.jupyter.display import display  # noqa
from syft_core import Client as SyftBoxClient  # noqa
from syft_rds.client.rds_client import RDSClient, init_session  # noqa
from syft_rds.client.setup import discover_rds_apps  # noqa



import atexit
import subprocess
import threading
import time
from pathlib import Path
from typing import Optional, Type, TypeVar
from uuid import UUID

from loguru import logger
from syft_core import Client as SyftBoxClient
from syft_event import SyftEvents

from syft_rds.client.client_registry import GlobalClientRegistry
from syft_rds.client.connection import get_connection
from syft_rds.client.local_store import LocalStore
from syft_rds.client.rds_clients.base import (
    RDSClientBase,
    RDSClientConfig,
    RDSClientModule,
)
from syft_rds.client.rds_clients.custom_function import CustomFunctionRDSClient
from syft_rds.client.rds_clients.dataset import DatasetRDSClient
from syft_rds.client.rds_clients.job import JobRDSClient
from syft_rds.client.rds_clients.runtime import RuntimeRDSClient
from syft_rds.client.rds_clients.user_code import UserCodeRDSClient
from syft_rds.client.rpc import RPCClient
from syft_rds.client.utils import PathLike, copy_dir_contents, deprecation_warning
from syft_rds.models import CustomFunction, Dataset, Job, JobStatus, Runtime, UserCode
from syft_rds.models.base import ItemBase
from syft_rds.server.app import create_app
from syft_rds.syft_runtime.main import (
    FileOutputHandler,
    JobConfig,
    RichConsoleUI,
    TextUI,
    get_runner_cls,
)
from syft_rds.utils.constants import JOB_STATUS_POLLING_INTERVAL

T = TypeVar("T", bound=ItemBase)

# Global registry to track running servers
_RUNNING_RDS_SERVERS = {}


def _is_server_running(host: str) -> bool:
    """Check if syft-rds server is running for the given host."""
    try:
        # Create a minimal config and connection to test server health
        config = RDSClientConfig(host=host)
        syftbox_client = _resolve_syftbox_client()
        connection = get_connection(syftbox_client, None, mock=False)
        rpc_client = RPCClient(config, connection)

        # Try health check with short timeout
        rpc_client.health(expiry="3s")
        return True
    except Exception:
        return False


def _start_server_thread(syftbox_client: SyftBoxClient) -> dict:
    """Start syft-rds server in a background thread."""
    rds_app: SyftEvents = create_app(client=syftbox_client)

    def run_server():
        try:
            logger.info(f"Starting syft-rds server for {syftbox_client.email}")
            rds_app.run_forever()
        except Exception as e:
            logger.error(f"Server thread failed: {e}")

    thread = threading.Thread(target=run_server, daemon=True)
    thread.start()

    return {"thread": thread, "server": rds_app}


def _wait_for_server(host: str, timeout: int = 30) -> bool:
    """Wait for server to be ready."""
    start_time = time.time()
    while time.time() - start_time < timeout:
        if _is_server_running(host):
            return True
        time.sleep(0.5)
    return False


def _ensure_server_running(
    syftbox_client: SyftBoxClient, auto_start: bool = False
) -> bool:
    """Ensure syft-rds server is running, starting it if needed."""
    host = syftbox_client.email

    if _is_server_running(host):
        return True

    if not auto_start:
        return False

    # Check if we already have a server thread for this client
    if host in _RUNNING_RDS_SERVERS and _RUNNING_RDS_SERVERS[host]["thread"].is_alive():
        # Server thread exists, wait for it to be ready
        if _wait_for_server(host):
            return True

    logger.info(f"syft-rds server not running for {host}, starting automatically...")

    # Start new server thread
    server_info = _start_server_thread(syftbox_client)
    _RUNNING_RDS_SERVERS[host] = server_info

    # Register cleanup on exit
    atexit.register(_cleanup_servers)

    # Wait for server to be ready
    if _wait_for_server(host):
        logger.success(f"syft-rds server is ready for {host}")
        return True
    else:
        logger.error(f"syft-rds server failed to start for {host}")
        return False


def _stop_server(host: str) -> bool:
    """Stop the syft-rds server for a specific host.

    Args:
        host: Email of the host whose server to stop

    Returns:
        True if server was stopped, False if no server was running
    """
    if host not in _RUNNING_RDS_SERVERS:
        return False

    server_info = _RUNNING_RDS_SERVERS[host]
    thread = server_info["thread"]
    server = server_info["server"]

    if thread.is_alive():
        logger.info(f"Stopping syft-rds server for {host}")
        try:
            server.stop()
            thread.join(timeout=5)  # Wait up to 5 seconds for clean shutdown
            if thread.is_alive():
                logger.warning(f"Server thread for {host} did not stop gracefully")
        except Exception as e:
            logger.error(f"Error stopping server for {host}: {e}")

    del _RUNNING_RDS_SERVERS[host]
    logger.success(f"Stopped syft-rds server for {host}")
    return True


def _cleanup_servers() -> None:
    """Clean up server threads on exit."""
    for host, server_info in _RUNNING_RDS_SERVERS.items():
        thread = server_info["thread"]
        server = server_info["server"]

        if thread.is_alive():
            logger.debug(f"Cleaning up server thread for {host}")
            try:
                server.stop()
            except Exception as e:
                logger.debug(f"Error during server cleanup for {host}: {e}")

    _RUNNING_RDS_SERVERS.clear()


def _resolve_syftbox_client(
    syftbox_client: Optional[SyftBoxClient] = None,
    config_path: Optional[PathLike] = None,
) -> SyftBoxClient:
    """
    Resolve a SyftBox client from either a provided instance or config path.

    Args:
        syftbox_client (SyftBoxClient, optional): Pre-configured client instance
        config_path (Union[str, Path], optional): Path to client config file

    Returns:
        SyftBoxClient: The SyftBox client instance

    Raises:
        ValueError: If both syftbox_client and config_path are provided
    """
    if (
        syftbox_client
        and config_path
        and syftbox_client.config_path.resolve() != Path(config_path).resolve()
    ):
        raise ValueError("Cannot provide both syftbox_client and config_path.")

    if syftbox_client:
        return syftbox_client

    return SyftBoxClient.load(filepath=config_path)


def init_session(
    host: str,
    syftbox_client: Optional[SyftBoxClient] = None,
    mock_server: Optional[SyftEvents] = None,
    syftbox_client_config_path: Optional[PathLike] = None,
    start_rds_server: bool = False,
    **config_kwargs,
) -> "RDSClient":
    """
    Initialize a session with the RDSClient.

    Args:
        host (str): The email of the remote datasite
        syftbox_client (SyftBoxClient, optional): Pre-configured SyftBox client instance.
            Takes precedence over syftbox_client_config_path.
        mock_server (SyftEvents, optional): Server for testing. If provided, uses
            a mock in-process RPC connection.
        syftbox_client_config_path (PathLike, optional): Path to client config file.
            Only used if syftbox_client is not provided.
        start_rds_server (bool, optional): Whether to automatically start syft-rds server
            if it's not running (same as `uv run syft-rds server`). Defaults to False.
        **config_kwargs: Additional configuration options for the RDSClient.

    Returns:
        RDSClient: The configured RDS client instance.
    """
    config = RDSClientConfig(host=host, **config_kwargs)
    syftbox_client = _resolve_syftbox_client(syftbox_client, syftbox_client_config_path)

    # Auto-start server if not using mock and auto_start_server is enabled
    use_mock = mock_server is not None
    if not use_mock and start_rds_server:
        server_started = _ensure_server_running(syftbox_client, start_rds_server)
        if not server_started:
            logger.warning(
                f"Failed to start syft-rds server for {host}. "
                "You may need to start it manually by running `uv run syft-rds server` in a separate terminal."
            )

    connection = get_connection(syftbox_client, mock_server, mock=use_mock)
    rpc_client = RPCClient(config, connection)
    local_store = LocalStore(config, syftbox_client)
    return RDSClient(config, rpc_client, local_store)


class RDSClient(RDSClientBase):
    def __init__(
        self,
        config: RDSClientConfig,
        rpc_client: RPCClient,
        local_store: LocalStore,
    ) -> None:
        super().__init__(config, rpc_client, local_store)
        self.job = JobRDSClient(self.config, self.rpc, self.local_store, parent=self)
        self.dataset = DatasetRDSClient(
            self.config, self.rpc, self.local_store, parent=self
        )
        self.user_code = UserCodeRDSClient(
            self.config, self.rpc, self.local_store, parent=self
        )
        self.custom_function = CustomFunctionRDSClient(
            self.config, self.rpc, self.local_store, parent=self
        )
        self.runtime = RuntimeRDSClient(
            self.config, self.rpc, self.local_store, parent=self
        )

        # GlobalClientRegistry is used to register this client, to enable referencing the client from returned objects
        # e.g. Job._client references the RDSClient instance that created it.
        GlobalClientRegistry.register_client(self)

        self._type_map: dict[Type[T], RDSClientModule[T]] = {
            Job: self.job,
            Dataset: self.dataset,
            Runtime: self.runtime,
            UserCode: self.user_code,
            CustomFunction: self.custom_function,
        }

        self.start()

    def __del__(self) -> None:
        self.close()

    def start(self) -> None:
        self._start_job_polling()

    def close(self) -> None:
        if self._polling_stop_event.is_set():
            return
        logger.debug("Stopping job polling thread.")
        self._polling_stop_event.set()
        self._polling_thread.join(timeout=2)

    def for_type(self, type_: Type[T]) -> RDSClientModule[T]:
        if type_ not in self._type_map:
            raise ValueError(f"No client registered for type {type_}")
        return self._type_map[type_]

    @property
    def uid(self) -> UUID:
        return self.config.uid

    @property
    @deprecation_warning(reason="client.jobs has been renamed to client.job")
    def jobs(self) -> JobRDSClient:
        return self.job

    @property
    @deprecation_warning(reason="Use client.dataset.get_all() instead.")
    def datasets(self) -> list[Dataset]:
        """Returns all available datasets.

        Returns:
            list[Dataset]: A list of all datasets
        """
        return self.dataset.get_all()

    # TODO move all logic under here to a separate job handler module

    def run_private(
        self,
        job: Job,
        display_type: str = "text",
        show_stdout: bool = True,
        show_stderr: bool = True,
        blocking: bool = True,
    ) -> Job:
        if job.status == JobStatus.rejected:
            raise ValueError(
                "Cannot run rejected job. "
                "If you want to override this, set `job.status` to something else."
            )
        logger.debug(f"Running job '{job.name}' on private data")
        job_config: JobConfig = self._get_config_for_job(job, blocking=blocking)
        result = self._run(
            job,
            job_config,
            display_type,
            show_stdout,
            show_stderr,
        )

        if isinstance(result, tuple):  # result from a blocking job
            return_code, error_message = result
            job_update = job.get_update_for_return_code(
                return_code=return_code, error_message=error_message
            )
            return self.job.update_job_status(job_update, job)
        else:  # non-blocking job
            return self._register_nonblocking_job(result, job)

    def run_mock(
        self,
        job: Job,
        display_type: str = "text",
        show_stdout: bool = True,
        show_stderr: bool = True,
        blocking: bool = True,
    ) -> Job:
        logger.debug(f"Running job '{job.name}' on mock data")
        job_config: JobConfig = self._get_config_for_job(job, blocking=blocking)
        job_config.data_path = self.dataset.get(name=job.dataset_name).get_mock_path()
        result = self._run(
            job,
            job_config,
            display_type,
            show_stdout,
            show_stderr,
        )
        logger.info(f"Result from running job '{job.name}' on mock data: {result}")
        return job

    def _get_config_for_job(self, job: Job, blocking: bool = True) -> JobConfig:
        user_code = self.user_code.get(job.user_code_id)
        dataset = self.dataset.get(name=job.dataset_name)
        runtime = self.runtime.get(job.runtime_id)
        runner_config = self.config.runner_config
        job_config = JobConfig(
            data_path=dataset.get_private_path(),
            function_folder=user_code.local_dir,
            runtime=runtime,
            args=[user_code.entrypoint],
            job_folder=runner_config.job_output_folder / job.uid.hex,
            timeout=runner_config.timeout,
            blocking=blocking,
        )
        return job_config

    def _prepare_job(self, job: Job, config: JobConfig) -> None:
        if job.custom_function_id is not None:
            self._prepare_custom_function(
                code_dir=job.user_code.local_dir,
                custom_function_id=job.custom_function_id,
            )

    def _prepare_custom_function(
        self,
        code_dir: Path,
        custom_function_id: UUID,
    ) -> None:
        custom_function = self.custom_function.get(uid=custom_function_id)

        try:
            copy_dir_contents(
                src=custom_function.local_dir,
                dst=code_dir,
                exists_ok=False,
            )
        except FileExistsError as e:
            raise FileExistsError(
                f"Cannot copy custom function files to {code_dir}: {e}"
            ) from e

    def _get_display_handler(
        self, display_type: str, show_stdout: bool, show_stderr: bool
    ):
        """Returns the appropriate display handler based on the display type."""
        if display_type == "rich":
            return RichConsoleUI(
                show_stdout=show_stdout,
                show_stderr=show_stderr,
            )
        elif display_type == "text":
            return TextUI(
                show_stdout=show_stdout,
                show_stderr=show_stderr,
            )
        else:
            raise ValueError(f"Unknown display type: {display_type}")

    def _run(
        self,
        job: Job,
        job_config: JobConfig,
        display_type: str = "text",
        show_stdout: bool = True,
        show_stderr: bool = True,
    ) -> int | subprocess.Popen:
        display_handler = self._get_display_handler(
            display_type, show_stdout, show_stderr
        )
        runner_cls = get_runner_cls(job_config)
        runner = runner_cls(
            handlers=[FileOutputHandler(), display_handler],
            update_job_status_callback=self.job.update_job_status,
        )

        self._prepare_job(job, job_config)
        return runner.run(job, job_config)

    def _start_job_polling(self) -> None:
        """Starts the job polling mechanism."""
        logger.debug("Starting thread to poll jobs.")
        self._non_blocking_jobs: dict[UUID, tuple[Job, subprocess.Popen]] = {}
        self._jobs_lock = threading.Lock()
        self._polling_stop_event = threading.Event()
        self._polling_thread = threading.Thread(
            target=self._poll_update_nonblocking_jobs
        )
        self._polling_thread.daemon = True
        self._polling_thread.start()

    def _register_nonblocking_job(self, result: subprocess.Popen, job: Job) -> Job:
        with self._jobs_lock:
            self._non_blocking_jobs[job.uid] = (job, result)
        logger.debug(f"Non-blocking job '{job.name}' started with PID {result.pid}")
        return job

    def _poll_update_nonblocking_jobs(self) -> None:
        """
        Polls for non-blocking jobs and updates the job status accordingly.
        If a job is finished, it is removed from the list of non-blocking jobs.
        """
        while not self._polling_stop_event.is_set():
            with self._jobs_lock:
                finished_jobs = []
                for job_uid, (job, process) in self._non_blocking_jobs.items():
                    if process.poll() is not None:
                        finished_jobs.append(job_uid)
                        try:
                            return_code = process.returncode
                            stderr = process.stderr.read() if process.stderr else None

                            # TODO: remove this once we have a better way to handle errors
                            if return_code == 0 and stderr and "| ERROR" in stderr:
                                logger.debug(
                                    "Error detected in logs, even with return code 0."
                                )
                                return_code = 1

                            job_update = job.get_update_for_return_code(
                                return_code=return_code, error_message=stderr
                            )
                            self.job.update_job_status(job_update, job)
                            logger.debug(
                                f"Non-blocking job '{job.name}' (PID: {process.pid}) "
                                f"finished with code {return_code}."
                            )
                        except Exception as e:
                            logger.error(
                                f"Error updating status for job {job.name}: {e}"
                            )

                for job_uid in finished_jobs:
                    del self._non_blocking_jobs[job_uid]

            time.sleep(JOB_STATUS_POLLING_INTERVAL)

    def stop_server(self) -> bool:
        """Stop the syft-rds server for this client's host.

        Returns:
            True if server was stopped, False if no server was running for this host
        """
        return _stop_server(self.config.host)



import os
from collections import UserDict, UserList

from syft_rds.display_utils.jupyter.tabulator import build_tabulator_table

SYFT_NO_REPR_HTML = "SYFT_NO_REPR_HTML" in os.environ


class TableList(UserList):
    def _repr_html_(self) -> str:
        if SYFT_NO_REPR_HTML:
            return None
        return build_tabulator_table(self.data)


class TableDict(UserDict):
    def _repr_html_(self) -> str:
        if SYFT_NO_REPR_HTML:
            return None
        return build_tabulator_table(self.data)



import functools
import os
import shutil
import warnings
from pathlib import Path
from typing import Iterator, TypeAlias, Union

PathLike: TypeAlias = Union[str, os.PathLike, Path]


def to_path(path: PathLike) -> Path:
    return Path(path).expanduser().resolve()


def deprecation_warning(reason: str = ""):
    """Decorator to mark functions as deprecated with a warning."""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warning_str = f"Function '{func.__name__}' is deprecated and will be removed in a future version."
            if reason:
                warning_str += f" {reason}"
            warnings.warn(
                warning_str,
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def copy_dir_contents(src: Path, dst: Path, exists_ok: bool = False) -> None:
    if not src.is_dir():
        raise ValueError(f"Source path {src} is not a directory.")
    return copy_paths(src.iterdir(), dst, exists_ok)


def copy_paths(files: Iterator[Path], dst: Path, exists_ok: bool = False) -> None:
    """
    Copy a list of files to a destination directory.
    If `dst` does not exist, it will be created.
    """
    if not dst.exists():
        dst.mkdir(parents=True, exist_ok=True)

    for file in files:
        dst_path = dst / file.name
        if dst_path.exists() and not exists_ok:
            raise FileExistsError(f"Destination path {dst_path} already exists.")
        if file.is_file():
            shutil.copy2(file, dst_path)
        elif file.is_dir():
            shutil.copytree(file, dst_path, dirs_exist_ok=exists_ok)



import jinja2
import pkg_resources


def render_data_access_template(filename):
    """
    Render the data access template with the given filename.

    Args:
        filename (str): The name of the file to access in the dataset

    Returns:
        str: The rendered template as a string
    """
    # Get the template path using pkg_resources
    template_path = pkg_resources.resource_filename("syft_rds", "assets")

    # Set up the Jinja environment
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_path),
        trim_blocks=True,
        lstrip_blocks=True,
    )

    # Load the template
    template = env.get_template("data_access_template.py.jinja")

    # Render the template with the given filename
    return template.render(filename=filename)



import os
import subprocess
import time
from pathlib import Path
from typing import Callable, Protocol, Type

from loguru import logger
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.spinner import Spinner

from syft_rds.models import (
    DockerMount,
    Job,
    JobConfig,
    JobUpdate,
    RuntimeKind,
)
from syft_rds.syft_runtime.mounts import get_mount_provider

DEFAULT_WORKDIR = "/app"
DEFAULT_OUTPUT_DIR = DEFAULT_WORKDIR + "/output"


class JobOutputHandler(Protocol):
    """Protocol defining the interface for job output handling and display"""

    def on_job_start(self, job_config: JobConfig) -> None:
        """Display job configuration"""
        pass

    def on_job_progress(self, stdout: str, stderr: str) -> None:
        """Display job progress"""
        pass

    def on_job_completion(self, return_code: int) -> None:
        """Display job completion status"""
        pass


class FileOutputHandler(JobOutputHandler):
    """Handles writing job output to log files"""

    def __init__(self):
        pass

    def on_job_start(self, job_config: JobConfig) -> None:
        self.config = job_config
        self.stdout_file = (job_config.logs_dir / "stdout.log").open("w")
        self.stderr_file = (job_config.logs_dir / "stderr.log").open("w")
        self.on_job_progress(stdout="Starting job...\n", stderr="Starting job...\n")

    def on_job_progress(self, stdout: str, stderr: str) -> None:
        if stdout:
            self.stdout_file.write(stdout)
            self.stdout_file.flush()
        if stderr:
            self.stderr_file.write(stderr)
            self.stderr_file.flush()

    def on_job_completion(self, return_code: int) -> None:
        self.on_job_progress(
            stdout=f"Job completed with return code {return_code}\n",
            stderr=f"Job completed with return code {return_code}\n",
        )
        self.close()

    def close(self) -> None:
        self.stdout_file.close()
        self.stderr_file.close()


# Helper function to limit path depth
def limit_path_depth(path: Path, max_depth: int = 4) -> str:
    parts = path.parts
    if len(parts) <= max_depth:
        return str(path)
    return str(Path("...") / Path(*parts[-max_depth:]))


class RichConsoleUI(JobOutputHandler):
    """Rich console implementation of JobOutputHandler"""

    def __init__(self, show_stdout: bool = True, show_stderr: bool = True):
        self.show_stdout = show_stdout
        self.show_stderr = show_stderr
        self.console = Console()
        spinner = Spinner("dots")
        self.live = Live(spinner, refresh_per_second=10)

    def on_job_start(self, job_config: JobConfig) -> None:
        self.console.print(
            Panel.fit(
                "\n".join(
                    [
                        "[bold green]Starting job[/]",
                        f"[bold white]Execution:[/] [cyan]{' '.join(job_config.runtime.cmd)} {' '.join(job_config.args)}[/]",
                        f"[bold white]Dataset Dir.:[/]  [cyan]{limit_path_depth(job_config.data_path)}[/]",
                        f"[bold white]Output Dir.:[/]   [cyan]{limit_path_depth(job_config.output_dir)}[/]",
                        f"[bold white]Timeout:[/]  [cyan]{job_config.timeout}s[/]",
                    ]
                ),
                title="[bold]Job Configuration",
                border_style="cyan",
            )
        )
        try:
            self.live.start()
            self.live.console.print("[bold cyan]Running job...[/]")
        except Exception as e:
            self.console.print(f"[red]Error starting live: {e}[/]")

    def on_job_progress(self, stdout: str, stderr: str) -> None:
        # Update UI display
        if not self.live:
            return

        if stdout and self.show_stdout:
            self.live.console.print(stdout, end="")
        if stderr and self.show_stderr:
            self.live.console.print(f"[red]{stderr}[/]", end="")

    def on_job_completion(self, return_code: int) -> None:
        # Update UI display
        if self.live:
            self.live.stop()

        if return_code == 0:
            self.console.print("\n[bold green]Job completed successfully![/]")
        else:
            self.console.print(
                f"\n[bold red]Job failed with return code {return_code}[/]"
            )

    def __del__(self):
        self.live.stop()


class TextUI(JobOutputHandler):
    """Simple text-based implementation of JobOutputHandler using print statements"""

    def __init__(self, show_stdout: bool = True, show_stderr: bool = True):
        self.show_stdout = show_stdout
        self.show_stderr = show_stderr
        self._job_running = False

    def on_job_start(self, config: JobConfig) -> None:
        first_line = "================ Job Configuration ================"
        last_line = "=" * len(first_line)
        print(f"\n{first_line}")
        print(f"Execution:    {' '.join(config.runtime.cmd)} {' '.join(config.args)}")
        print(f"Dataset Dir.: {limit_path_depth(config.data_path)}")
        print(f"Output Dir.:  {limit_path_depth(config.output_dir)}")
        print(f"Timeout:      {config.timeout}s")
        print(f"{last_line}\n")
        print("[STARTING JOB]")
        self._job_running = True

    def on_job_progress(self, stdout: str, stderr: str) -> None:
        if not self._job_running:
            return
        if stdout and self.show_stdout:
            print(stdout, end="")
        if stderr and self.show_stderr:
            print(f"[STDERR] {stderr}", end="")

    def on_job_completion(self, return_code: int) -> None:
        self._job_running = False
        if return_code == 0:
            print("\n[JOB COMPLETED SUCCESSFULLY]\n")
        else:
            print(f"\n[JOB FAILED] Return code: {return_code}\n")

    def __del__(self):
        self._job_running = False


class JobRunner:
    """Base class for running jobs."""

    def __init__(
        self,
        handlers: list[JobOutputHandler],
        update_job_status_callback: Callable[[JobUpdate, Job], Job | None],
    ):
        self.handlers = handlers
        self.update_job_status_callback = update_job_status_callback

    def run(
        self,
        job: Job,
        job_config: JobConfig,
    ) -> tuple[int, str | None] | subprocess.Popen:
        """Run a job
        Returns:
            tuple[int, str | None]: (blocking mode) The return code and error message
                if the job failed, otherwise None.
            subprocess.Popen: (non-blocking mode) The process object.
        """
        raise NotImplementedError

    def _prepare_job_folders(self, job_config: JobConfig) -> None:
        """Create necessary job folders"""
        job_config.job_path.mkdir(parents=True, exist_ok=True)
        job_config.logs_dir.mkdir(exist_ok=True)
        job_config.output_dir.mkdir(exist_ok=True)
        os.chmod(job_config.output_dir, 0o777)

    def _validate_paths(self, job_config: JobConfig) -> None:
        """Validate that the necessary paths exist and are of the correct type."""
        if not job_config.function_folder.exists():
            raise ValueError(
                f"Function folder {job_config.function_folder} does not exist"
            )
        if not job_config.data_path.exists():
            raise ValueError(f"Dataset folder {job_config.data_path} does not exist")

    def _run_subprocess(
        self,
        cmd: list[str],
        job_config: JobConfig,
        job: Job,
        env: dict | None = None,
        blocking: bool = True,
    ) -> tuple[int, str | None] | subprocess.Popen:
        """
        Returns:
            tuple[int, str | None]: (blocking mode) The return code and error message
                if the job failed, otherwise None.
            subprocess.Popen: (non-blocking mode) The process object.
        """
        if self.update_job_status_callback:
            job_update = job.get_update_for_in_progress()
            self.update_job_status_callback(job_update, job)

        for handler in self.handlers:
            handler.on_job_start(job_config)

        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            env=env,
        )

        if blocking:
            logger.info("Running job in blocking mode")
            return self._run_blocking(process, job)
        else:
            logger.info("Running job in non-blocking mode")
            return process

    def _run_blocking(
        self,
        process: subprocess.Popen,
        job: Job,
    ) -> tuple[int, str | None]:
        stderr_logs = []

        # Stream logs
        while True:
            stdout_line = process.stdout.readline()
            stderr_line = process.stderr.readline()
            if stderr_line:
                stderr_logs.append(stderr_line)
            if stdout_line or stderr_line:
                for handler in self.handlers:
                    handler.on_job_progress(stdout_line, stderr_line)
            if process.poll() is not None:
                logger.debug(
                    f"Process {process.pid} terminated with return code {process.returncode}"
                )
                break
            time.sleep(0.1)

        # Flush remaining output
        for line in process.stdout:
            for handler in self.handlers:
                handler.on_job_progress(line, "")
        for line in process.stderr:
            stderr_logs.append(line)
            for handler in self.handlers:
                handler.on_job_progress("", line)

        return_code = process.returncode
        logger.debug(f"Return code: {return_code}")
        error_message = None
        if stderr_logs:
            logger.debug(f"Stderr logs: {stderr_logs}")
            error_message = "\n".join(stderr_logs)

            # TODO: remove this once we have a better way to handle errors
            if return_code == 0 and error_message and "| ERROR" in error_message:
                logger.debug("Error detected in logs, even with return code 0.")
                return_code = 1

        # Handle job completion results
        for handler in self.handlers:
            handler.on_job_completion(process.returncode)

        return return_code, error_message


class PythonRunner(JobRunner):
    """Runs a Python job in a local subprocess."""

    def run(
        self,
        job: Job,
        job_config: JobConfig,
    ) -> tuple[int, str | None] | subprocess.Popen:
        """Run a job"""
        self._validate_paths(job_config)
        self._prepare_job_folders(job_config)

        cmd = self._prepare_run_command(job_config)

        env = os.environ.copy()
        env.update(job_config.get_env())
        env.update(job_config.extra_env)

        return self._run_subprocess(
            cmd, job_config, job, env=env, blocking=job_config.blocking
        )

    def _prepare_run_command(self, job_config: JobConfig) -> list[str]:
        return [
            *job_config.runtime.cmd,
            str(Path(job_config.function_folder) / job_config.args[0]),
            *job_config.args[1:],
        ]


class DockerRunner(JobRunner):
    """Runs a job in a Docker container."""

    def run(
        self,
        job: Job,
        job_config: JobConfig,
    ) -> tuple[int, str | None] | subprocess.Popen:
        """Run a job in a Docker container"""
        logger.debug(
            f"Running code in '{job_config.function_folder}' on dataset '{job_config.data_path}' with runtime '{job_config.runtime.kind.value}'"
        )

        self._validate_paths(job_config)
        self._prepare_job_folders(job_config)

        self._check_docker_daemon(job)
        self._check_or_build_image(job_config, job)

        cmd = self._prepare_run_command(job_config)

        return self._run_subprocess(cmd, job_config, job, blocking=job_config.blocking)

    def _check_docker_daemon(self, job: Job) -> None:
        """Check if the Docker daemon is running."""
        try:
            process = subprocess.run(
                ["docker", "info"],
                check=True,
                capture_output=True,
            )
        except Exception as e:
            if self.update_job_status_callback:
                job_update = job.get_update_for_return_code(
                    return_code=process.returncode,
                    error_message="Docker daemon is not running with error: " + str(e),
                )
                self.update_job_status_callback(job_update, job)
            raise RuntimeError("Docker daemon is not running with error: " + str(e))

    def _get_image_name(self, job_config: JobConfig) -> str:
        """Get the Docker image name from the config or use the default."""
        runtime_config = job_config.runtime.config
        if not runtime_config.image_name:
            return job_config.runtime.name
        return runtime_config.image_name

    def _check_or_build_image(self, job_config: JobConfig, job: Job) -> None:
        """Check if the Docker image exists, otherwise build it."""
        image_name = self._get_image_name(job_config)
        result = subprocess.run(
            ["docker", "image", "inspect", image_name],
            capture_output=True,
            check=False,
            text=True,
        )
        if result.returncode == 0:
            logger.info(f"Docker image '{image_name}' already exists.")
            return

        logger.info(f"Docker image '{image_name}' not found. Building it now...")
        self._build_docker_image(job_config, job)

    def _build_docker_image(self, job_config: JobConfig, job: Job) -> None:
        """Build the Docker image."""
        image_name = self._get_image_name(job_config)
        dockerfile_content: str = job_config.runtime.config.dockerfile_content
        error_for_job: str | None = None
        build_context = "."
        try:
            build_cmd = [
                "docker",
                "build",
                "-t",
                image_name,
                "-f",
                "-",  # Use stdin for Dockerfile content
                str(build_context),
            ]
            logger.debug(
                f"Running docker build command: {' '.join(build_cmd)}\nDockerfile content:\n{dockerfile_content}"
            )
            process = subprocess.run(
                build_cmd,
                input=dockerfile_content,
                capture_output=True,
                check=True,
                text=True,
            )

            logger.debug(process.stdout)
            logger.info(f"Successfully built Docker image '{image_name}'.")
        except FileNotFoundError:
            raise RuntimeError("Docker not installed or not in PATH.")
        except subprocess.CalledProcessError as e:
            error_message = f"Failed to build Docker image '{image_name}'."
            logger.error(f"{error_message} stderr: {e.stderr}")
            error_for_job = f"{error_message}\n{e.stderr}"
            raise RuntimeError(f"Failed to build Docker image '{image_name}'.")
        except Exception as e:
            raise RuntimeError(f"An error occurred during Docker image build: {e}")
        finally:
            if error_for_job and self.update_job_status_callback:
                job_failed = job.get_update_for_return_code(
                    return_code=process.returncode,
                    error_message=error_for_job,
                )
                self.update_job_status_callback(job_failed, job)

    def _get_extra_mounts(self, job_config: JobConfig) -> list[DockerMount]:
        """Get extra mounts for a job"""
        docker_runtime_config = job_config.runtime.config
        if docker_runtime_config.app_name is None:
            return []
        mount_provider = get_mount_provider(docker_runtime_config.app_name)
        if mount_provider:
            return mount_provider.get_mounts(job_config)
        return []

    def _prepare_run_command(self, job_config: JobConfig) -> list[str]:
        """Build the Docker run command with security constraints"""
        image_name = self._get_image_name(job_config)
        docker_mounts = [
            "-v",
            f"{Path(job_config.function_folder).absolute()}:{DEFAULT_WORKDIR}/code:ro",
            "-v",
            f"{Path(job_config.data_path).absolute()}:{DEFAULT_WORKDIR}/data:ro",
            "-v",
            f"{job_config.output_dir.absolute()}:{DEFAULT_OUTPUT_DIR}:rw",
        ]

        extra_mounts = self._get_extra_mounts(job_config)
        if extra_mounts:
            for mount in extra_mounts:
                docker_mounts.extend(
                    [
                        "-v",
                        f"{mount.source.resolve()}:{mount.target}:{mount.mode}",
                    ]
                )

        interpreter = " ".join(job_config.runtime.cmd)
        interpreter_str = f'"{interpreter}"' if " " in interpreter else interpreter

        limits = [
            # Security constraints
            "--cap-drop",
            "ALL",  # Drop all capabilities
            "--network",
            "none",  # Disable networking
            # "--read-only",  # Read-only root filesystem - TODO: re-enable this
            "--tmpfs",
            "/tmp:size=16m,noexec,nosuid,nodev",  # Secure temp directory
            # Resource limits
            "--memory",
            "1G",
            "--cpus",
            "1",
            "--pids-limit",
            "100",
            "--ulimit",
            "nproc=4096:4096",
            "--ulimit",
            "nofile=50:50",
            "--ulimit",
            "fsize=10000000:10000000",  # ~10MB file size limit
        ]

        docker_run_cmd = [
            "docker",
            "run",
            "--rm",  # Remove container after completion
            *limits,
            # Environment variables
            "-e",
            f"TIMEOUT={job_config.timeout}",
            "-e",
            f"DATA_DIR={job_config.data_mount_dir}",
            "-e",
            f"OUTPUT_DIR={DEFAULT_OUTPUT_DIR}",
            "-e",
            f"INTERPRETER={interpreter_str}",
            "-e",
            f"INPUT_FILE='{DEFAULT_WORKDIR}/code/{job_config.args[0]}'",
            *job_config.get_extra_env_as_docker_args(),
            *docker_mounts,
            "--workdir",
            DEFAULT_WORKDIR,
            image_name,
            f"{DEFAULT_WORKDIR}/code/{job_config.args[0]}",
            *job_config.args[1:],
        ]
        logger.debug(f"Docker run command: {docker_run_cmd}")
        return docker_run_cmd


def get_runner_cls(job_config: JobConfig) -> Type[JobRunner]:
    """Factory to get the appropriate runner class for a job config."""
    runtime_kind = job_config.runtime.kind
    if runtime_kind == RuntimeKind.PYTHON:
        return PythonRunner
    elif runtime_kind == RuntimeKind.DOCKER:
        return DockerRunner
    else:
        raise NotImplementedError(f"Unsupported runtime kind: {runtime_kind}")



from pathlib import Path
from typing import Optional, Union
from uuid import UUID

from syft_rds.client.rds_clients.base import RDSClientModule
from syft_rds.client.rds_clients.utils import ensure_is_admin
from syft_rds.models import (
    Dataset,
    DatasetCreate,
    DatasetUpdate,
)


class DatasetRDSClient(RDSClientModule[Dataset]):
    ITEM_TYPE = Dataset

    @ensure_is_admin
    def create(
        self,
        name: str,
        path: Union[str, Path],
        mock_path: Union[str, Path],
        summary: Optional[str] = None,
        description_path: Optional[Union[str, Path]] = None,
        tags: list[str] = [],
        runtime_id: Optional[UUID] = None,
    ) -> Dataset:
        dataset_create = DatasetCreate(
            name=name,
            path=str(path),
            mock_path=str(mock_path),
            summary=summary,
            description_path=str(description_path) if description_path else None,
            tags=tags,
            runtime_id=runtime_id,
        )
        return self.local_store.dataset.create(dataset_create)

    @ensure_is_admin
    def delete(self, name: str) -> bool:
        """Delete a dataset by name.

        Args:
            name: Name of the dataset to delete

        Returns:
            True if deletion was successful, False otherwise

        Raises:
            RuntimeError: If deletion fails due to file system errors
        """
        return self.local_store.dataset.delete_by_name(name)

    @ensure_is_admin
    def update(self, dataset_update: DatasetUpdate) -> Dataset:
        raise NotImplementedError("Dataset update is not supported yet")



from syft_rds.models.api_models import *  # noqa: F403
from syft_rds.models.custom_function_models import *  # noqa: F403
from syft_rds.models.dataset_models import *  # noqa: F403
from syft_rds.models.job_models import *  # noqa: F403
from syft_rds.models.runtime_models import *  # noqa: F403
from syft_rds.models.user_code_models import *  # noqa: F403



import json
import shutil
import tempfile
from pathlib import Path
from typing import Any
from uuid import UUID

from loguru import logger

from syft_rds.client.exceptions import RDSValidationError
from syft_rds.client.rds_clients.base import RDSClientModule
from syft_rds.client.utils import PathLike
from syft_rds.models import (
    Job,
    JobCreate,
    JobStatus,
    JobUpdate,
    Runtime,
    UserCode,
)
from syft_rds.models.custom_function_models import CustomFunction
from syft_rds.models.job_models import JobErrorKind, JobResults


class JobRDSClient(RDSClientModule[Job]):
    ITEM_TYPE = Job

    def submit(
        self,
        user_code_path: PathLike,
        dataset_name: str,
        entrypoint: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        custom_function: CustomFunction | UUID | None = None,
        runtime_name: str | None = None,
        runtime_kind: str | None = None,
        runtime_config: dict | None = None,
    ) -> Job:
        """`submit` is a convenience method to create both a UserCode and a Job in one call."""
        if custom_function is not None:
            custom_function_id = self._resolve_custom_func_id(custom_function)
            custom_function = (
                self.rds.custom_function.get(uid=custom_function_id)
                if custom_function_id
                else None
            )
            if entrypoint is not None:
                raise RDSValidationError(
                    "Cannot specify entrypoint when using a custom function."
                )
            entrypoint = custom_function.entrypoint

        user_code = self.rds.user_code.create(
            code_path=user_code_path, entrypoint=entrypoint
        )

        runtime = self.rds.runtime.create(
            runtime_name=runtime_name,
            runtime_kind=runtime_kind,
            config=runtime_config,
        )

        job = self.create(
            name=name,
            description=description,
            user_code=user_code,
            dataset_name=dataset_name,
            tags=tags,
            custom_function=custom_function,
            runtime=runtime,
        )

        return job

    def submit_with_params(
        self,
        dataset_name: str,
        custom_function: CustomFunction | UUID,
        **params: Any,
    ) -> Job:
        """
        Utility method to a job with parameters for a custom function.

        Args:
            dataset_name (str): The name of the dataset to use.
            custom_function (CustomFunction | UUID): The custom function to use.
            **params: Additional parameters to pass to the custom function.

        Returns:
            Job: The created job.
        """
        if isinstance(custom_function, UUID):
            custom_function = self.rds.custom_function.get(uid=custom_function)
        elif not isinstance(custom_function, CustomFunction):
            raise ValueError(
                f"Invalid custom_function type {type(custom_function)}. Must be CustomFunction or UUID"
            )

        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_dir_path = Path(tmpdir)
            user_params_path = tmp_dir_path / custom_function.input_params_filename
            if not user_params_path.suffix == ".json":
                raise ValueError(
                    f"Input params file must be a JSON file, got {user_params_path.suffix}. Please contact the administrator."
                )

            try:
                params_json = json.dumps(params)
            except Exception as e:
                raise ValueError(f"Failed to serialize params to JSON: {e}.") from e

            user_params_path.write_text(params_json)

            return self.submit(
                user_code_path=user_params_path,
                dataset_name=dataset_name,
                custom_function=custom_function,
            )

    def _resolve_custom_func_id(
        self, custom_function: CustomFunction | UUID | None
    ) -> UUID | None:
        if custom_function is None:
            return None
        if isinstance(custom_function, UUID):
            return custom_function
        elif isinstance(custom_function, CustomFunction):
            return custom_function.uid
        else:
            raise RDSValidationError(
                f"Invalid custom_function type {type(custom_function)}. Must be CustomFunction, UUID, or None"
            )

    def _resolve_usercode_id(self, user_code: UserCode | UUID) -> UUID:
        if isinstance(user_code, UUID):
            return user_code
        elif isinstance(user_code, UserCode):
            return user_code.uid
        else:
            raise RDSValidationError(
                f"Invalid user_code type {type(user_code)}. Must be UserCode, UUID, or str"
            )

    def create(
        self,
        user_code: UserCode | UUID,
        dataset_name: str,
        runtime: Runtime,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        custom_function: CustomFunction | UUID | None = None,
    ) -> Job:
        # TODO ref dataset by UID instead of name
        user_code_id = self._resolve_usercode_id(user_code)
        custom_function_id = self._resolve_custom_func_id(custom_function)

        job_create = JobCreate(
            name=name,
            description=description,
            tags=tags if tags is not None else [],
            user_code_id=user_code_id,
            runtime_id=runtime.uid,
            dataset_name=dataset_name,
            custom_function_id=custom_function_id,
        )
        job = self.rpc.job.create(job_create)

        return job

    def _get_results_from_dir(
        self,
        job: Job,
        results_dir: PathLike,
    ) -> JobResults:
        """Get the job results from the specified directory, and format it into a JobResults object."""
        results_dir = Path(results_dir)
        if not results_dir.exists():
            raise ValueError(
                f"Results directory {results_dir} does not exist for job {job.uid}"
            )

        output_dir = results_dir / "output"
        logs_dir = results_dir / "logs"
        expected_layout_msg = (
            f"{results_dir} should contain 'output' and 'logs' directories."
        )
        if not output_dir.exists():
            raise ValueError(
                f"Output directory {output_dir.name} does not exist for job {job.uid}. "
                + expected_layout_msg
            )
        if not logs_dir.exists():
            raise ValueError(
                f"Logs directory {logs_dir.name} does not exist for job {job.uid}. "
                + expected_layout_msg
            )

        return JobResults(
            job=job,
            results_dir=results_dir,
        )

    def review_results(
        self, job: Job, output_dir: PathLike | None = None
    ) -> JobResults:
        if output_dir is None:
            output_dir = self.config.runner_config.job_output_folder / job.uid.hex
        return self._get_results_from_dir(job, output_dir)

    def share_results(self, job: Job) -> None:
        if not self.is_admin:
            raise RDSValidationError("Only admins can share results")
        job_results_folder = self.config.runner_config.job_output_folder / job.uid.hex
        output_path = self._share_result_files(job, job_results_folder)
        updated_job = self.rpc.job.update(
            JobUpdate(
                uid=job.uid,
                status=JobStatus.shared,
                error=job.error,
            )
        )
        job.apply_update(updated_job, in_place=True)
        logger.info(f"Shared results for job {job.uid} at {output_path}")

    def _share_result_files(self, job: Job, job_results_folder: Path) -> Path:
        syftbox_output_path = job.output_url.to_local_path(
            self.rds._syftbox_client.datasites
        )
        if not syftbox_output_path.exists():
            syftbox_output_path.mkdir(parents=True)

        # Copy all contents from job_output_folder to the output path
        for item in job_results_folder.iterdir():
            if item.is_file():
                shutil.copy2(item, syftbox_output_path)
            elif item.is_dir():
                shutil.copytree(
                    item,
                    syftbox_output_path / item.name,
                    dirs_exist_ok=True,
                )

        return syftbox_output_path

    def get_results(self, job: Job) -> JobResults:
        """Get the shared job results."""
        if job.status != JobStatus.shared:
            raise RDSValidationError(
                f"Job {job.uid} is not shared. Current status: {job.status}"
            )
        return self._get_results_from_dir(job, job.output_path)

    def reject(self, job: Job, reason: str = "Unspecified") -> None:
        if not self.is_admin:
            raise RDSValidationError("Only admins can reject jobs")

        allowed_statuses = (
            JobStatus.pending_code_review,
            JobStatus.job_run_finished,
            JobStatus.job_run_failed,
        )
        if self.status not in allowed_statuses:
            raise ValueError(f"Cannot reject job with status: {self.status}")

        error = (
            JobErrorKind.failed_code_review
            if job.status == JobStatus.pending_code_review
            else JobErrorKind.failed_output_review
        )

        job_update = JobUpdate(
            uid=job.uid,
            status=JobStatus.rejected,
            error=error,
            error_message=reason,
        )

        updated_job = self.rpc.job.update(job_update)
        job.apply_update(updated_job, in_place=True)

    def update_job_status(self, job_update: JobUpdate, job: Job) -> Job:
        new_job = self.rpc.job.update(job_update)
        return job.apply_update(new_job)

    def delete(self, job: Job | UUID, delete_orphaned_usercode: bool = True) -> bool:
        """Delete a single job by Job object or UUID.

        Args:
            job: Job object or UUID of the job to delete
            delete_orphaned_usercode: If True, also delete UserCode if not used by other jobs

        Returns:
            True if deletion was successful

        Raises:
            RDSValidationError: If user is not admin
        """
        if not self.is_admin:
            raise RDSValidationError("Only admins can delete jobs")

        # Get the full job object if we only have UUID
        if isinstance(job, UUID):
            try:
                job = self.get(uid=job, mode="local")
            except ValueError:
                logger.warning(f"Job {job} not found for deletion")
                return False

        # Delete job output folders
        self._delete_job_outputs(job)

        # Delete Job YAML file from local store
        deleted = self.local_store.job.delete_by_id(job.uid)
        if not deleted:
            logger.warning(f"Job {job.uid} not found for deletion")
            return False

        logger.info(f"Deleted job {job.uid} successfully")

        # Conditionally delete orphaned UserCode
        if delete_orphaned_usercode and job.user_code_id:
            self._delete_orphaned_usercode(job.user_code_id, job.uid)

        return True

    def delete_all(self, delete_orphaned_usercode: bool = True, **filters) -> int:
        """Delete all jobs matching the given filters.

        Args:
            delete_orphaned_usercode: If True, also delete UserCode if not used by other jobs
            **filters: Filter criteria for jobs to delete (e.g., status=JobStatus.rejected)

        Returns:
            Number of jobs deleted

        Raises:
            RDSValidationError: If user is not admin
        """
        if not self.is_admin:
            raise RDSValidationError("Only admins can delete jobs")

        # Get all jobs matching the filters
        jobs_to_delete = self.get_all(mode="local", **filters)

        deleted_count = 0
        for job in jobs_to_delete:
            if self.delete(job, delete_orphaned_usercode=delete_orphaned_usercode):
                deleted_count += 1

        logger.info(f"Deleted {deleted_count} jobs out of {len(jobs_to_delete)} found")
        return deleted_count

    def _delete_job_outputs(self, job: Job) -> None:
        """Delete job output folders."""
        # Delete job output folder using the job's output_url
        if job.output_url:
            job_output_path = job.output_url.to_local_path(
                self._syftbox_client.datasites
            )
            if job_output_path.exists():
                shutil.rmtree(job_output_path)
                logger.debug(f"Deleted job output path: {job_output_path}")

        # Delete job results from runner output folder
        job_runner_output = self.config.runner_config.job_output_folder / job.uid.hex
        if job_runner_output.exists():
            shutil.rmtree(job_runner_output)
            logger.debug(f"Deleted job runner output: {job_runner_output}")

    def _delete_orphaned_usercode(
        self, user_code_id: UUID, excluded_job_uid: UUID
    ) -> None:
        """Delete UserCode if it's not used by any other jobs."""
        # Check if UserCode is used by other jobs
        other_jobs = [
            j
            for j in self.get_all(mode="local", user_code_id=user_code_id)
            if j.uid != excluded_job_uid
        ]

        if other_jobs:
            logger.debug(
                f"UserCode {user_code_id} is still used by {len(other_jobs)} other job(s)"
            )
            return

        # UserCode is orphaned, delete it
        try:
            usercode = self.rds.user_code.get(uid=user_code_id, mode="local")

            # Delete UserCode files
            if usercode.dir_url:
                usercode_path = usercode.dir_url.to_local_path(
                    self._syftbox_client.datasites
                )
                if usercode_path.exists():
                    shutil.rmtree(usercode_path)
                    logger.debug(f"Deleted UserCode folder: {usercode_path}")

            # Delete UserCode YAML
            if self.local_store.user_code.delete_by_id(user_code_id):
                logger.debug(f"Deleted orphaned UserCode {user_code_id}")
        except Exception as e:
            logger.warning(f"Failed to delete orphaned UserCode {user_code_id}: {e}")



from functools import wraps
from pathlib import Path
from typing import Generic, Optional, Type, TypeVar
from uuid import UUID

import yaml
from pydantic import TypeAdapter

from syft_rds.models.base import ItemBase

PERMS = """
rules:
- pattern: '**'
  access:
    read:
    - '*'
"""

T = TypeVar("T", bound=ItemBase)


def ensure_store_exists(func):
    @wraps(func)
    def wrapper(self: "YAMLStore", *args, **kwargs):
        if not self.item_type_dir.exists():
            self.item_type_dir.mkdir(parents=True, exist_ok=True)
            perms_file = self.item_type_dir.parent / "syft.pub.yaml"
            perms_file.write_text(PERMS)  # TODO create more restrictive permissions
        return func(self, *args, **kwargs)

    return wrapper


class YAMLStore(Generic[T]):
    def __init__(self, item_type: Type[T], store_dir: str | Path):
        """A lightweight file-based database that stores records as individual YAML files.

        YAMLStore provides a simple database implementation where each record
        is stored as a separate YAML file in the filesystem. It supports basic CRUD operations,
        querying, and searching capabilities while using Pydantic models for  validation.

        The database creates a hierarchical directory structure:

        /store_dir/
        ├── model1_name/           # Directory for first model type
        │   ├── uuid1.yaml             # Individual record files
        │   └── uuid2.yaml
        ├── model2_name/           # Directory for second model type
        │   ├── uuid3.yaml
        │   └── uuid4.yaml
        └── syftperm.yaml              # Permissions file

        Where:
        - Each record is stored as a separate .yaml file
        - Filenames are UUIDs (e.g., "123e4567-e89b-12d3-a456-426614174000.yaml")
        - All files for a specific model are stored in a dedicated subdirectory named after the model's __schema_name__
        - A syftperm.yaml file is created at the parent level to manage permissions

        Features:
        - CRUD operations (Create, Read, Update, Delete)
        - Query records with exact field matching
        - Case-insensitive search across specified fields
        - Automatic UUID generation for new records
        - Type safety and validation through Pydantic models

        Example:
            ```python
            from pydantic import BaseModel

            class User(BaseModel):
                __schema_name__ = "users"
                name: str
                email: str

            # Initialize the database
            store = YAMLStore(User, "/path/to/store")

            # Create a new user
            user = User(name="John Doe", email="john@example.com")
            user_id = store.create(user)

            # Query users
            johns = store.get_all(filters={"name": "John Doe"})
            ```

        Args:
            schema: The Pydantic model class that defines the schema for stored records.
                Must inherit from ItemBase.
            store_dir: Directory path where the database files will be stored.
                    Can be string or Path object.

        Notes:
            - The database automatically creates the necessary directory structure
            - Each model type gets its own subdirectory based on __schema_name__
            - Records must be instances of Pydantic models inheriting from ItemBase
            - All operations are file-system based for now (no in-memory caching)
            - Suitable for smaller datasets where simple CRUD operations are needed
            - Provides human-readable storage format
        """
        self.item_type = item_type
        self.store_dir = Path(store_dir)
        self._field_validators = self._make_field_validators()

    def _make_field_validators(self) -> dict:
        """
        Create a dictionary of field_name: TypeAdapter for each field in the schema.
        These can be used to validate and convert field values to the correct type, required when querying the store.
        """
        return {
            field_name: TypeAdapter(field_info.annotation)
            for field_name, field_info in self.item_type.model_fields.items()
        }

    def _coerce_field_types(self, filters: dict) -> dict:
        """
        If possible, convert filter values to the correct type for the schema.
        e.g. convert str to UUID, or str to Enum, etc.
        """
        # TODO move filter type coercion to YAMLStore to avoid duplication serverside
        resolved_filters = {}
        for filter_name, filter_value in filters.items():
            validator = self._field_validators.get(filter_name, None)
            if validator is None:
                # Cannot infer type, leave it in the original form
                resolved_filters[filter_name] = filter_value
                continue
            try:
                type_adapter = self._field_validators[filter_name]
                validated_value = type_adapter.validate_python(filter_value)
                resolved_filters[filter_name] = validated_value
            except Exception:
                # Cannot convert to the correct type, leave it in the original form
                # logger.exception(
                #     f"Could not convert filter value {filter_value} to {field_info.annotation} for field {filter_name}"
                # )
                resolved_filters[filter_name] = filter_value
        return resolved_filters

    @property
    def item_type_dir(self) -> Path:
        return self.store_dir / self.item_type.__schema_name__

    def _get_record_path(self, uid: str | UUID) -> Path:
        """Get the full path for a record's YAML file from its UID."""
        return self.item_type_dir / f"{uid}.yaml"

    def _save_record(self, record: T) -> None:
        """Save a single record to its own YAML file"""
        file_path = self._get_record_path(record.uid)
        yaml_dump = yaml.safe_dump(
            record.model_dump(mode="json"),
            indent=2,
            sort_keys=False,
        )
        file_path.write_text(yaml_dump)

    @ensure_store_exists
    def get_by_uid(self, uid: str | UUID) -> Optional[T]:
        """Get a single record by UID"""
        file_path = self._get_record_path(uid)
        if not file_path.exists():
            return None
        record_dict = yaml.safe_load(file_path.read_text())
        return self.item_type.model_validate(record_dict)

    @ensure_store_exists
    def list_all(self) -> list[T]:
        """List all records in the store"""
        records = []
        for file_path in self.item_type_dir.glob("*.yaml"):
            _id = file_path.stem
            loaded_record = self.get_by_uid(_id)
            if loaded_record is not None:
                records.append(loaded_record)
        return records

    @ensure_store_exists
    def create(self, record: T, overwrite: bool = False) -> T:
        """
        Create a new record in the store

        Args:
            record: Instance of the model to create
            overwrite: If True, overwrite the record if it already exists

        Returns:
            UID of the created record
        """
        if not isinstance(record, self.item_type):
            raise TypeError(f"`record` must be of type {self.item_type.__name__}")
        file_path = self._get_record_path(record.uid)
        if file_path.exists() and not overwrite:
            raise ValueError(f"Record with UID {record.uid} already exists")
        self._save_record(record)
        return record

    @ensure_store_exists
    def update(self, uid: str | UUID, record: T) -> Optional[T]:
        """
        Update a record by UID

        Args:
            uid: Record UID to update
            record: New data to update with

        Returns:
            Updated record if found, None otherwise
        """
        if not isinstance(record, self.item_type):
            raise TypeError(f"`record` must be of type {self.item_type.__name__}")

        existing_record = self.get_by_uid(uid)
        if not existing_record:
            return None

        # Update the record
        updated_record = existing_record.model_copy(
            update=record.model_dump(exclude={"uid"})
        )
        self._save_record(updated_record)
        return updated_record

    @ensure_store_exists
    def delete(self, uid: str | UUID) -> bool:
        """
        Delete a record by UID

        Args:
            uid: Record UID to delete

        Returns:
            True if record was deleted, False if not found
        """
        file_path = self._get_record_path(uid)
        if not file_path.exists():
            return False
        file_path.unlink()
        return True

    @ensure_store_exists
    def get_one(self, **filters) -> Optional[T]:
        """
        Get one record with exact match filters.

        Args:
            **filters: Field-value pairs to filter by

        Returns:
            Matching record or None
        """
        if len(filters.keys()) == 1 and "uid" in filters:
            return self.get_by_uid(filters["uid"])

        else:
            res = self.get_all(filters=filters, limit=1)
            if len(res) == 0:
                return None
            return res[0]

    def _sort_items(self, items: list[T], order_by: str, sort_order: str) -> list[T]:
        return sorted(
            items,
            key=lambda x: getattr(x, order_by, None),
            reverse=sort_order == "desc",
        )

    @ensure_store_exists
    def get_all(
        self,
        limit: Optional[int] = None,
        offset: int = 0,
        order_by: Optional[str] = None,
        sort_order: str = "asc",
        filters: Optional[dict] = None,
    ) -> list[T]:
        """
        Get all records with optional filtering, sorting, and pagination.
        Filters are case-sensitive and must match exactly.

        Args:
            limit (Optional[int], optional): limit. Defaults to None.
            offset (int, optional): offset. Defaults to 0.
            order_by (Optional[str], optional): field to order by. Defaults to None.
            sort_order (str, optional): 'asc' or 'desc'. Defaults to "asc".
            filters (Optional[dict], optional): dictionary of filters, Pydantic is used for type coercion,
                so comparing strings to UUIDs or dates will work. Defaults to None.

        Returns:
            list[T]: List of matching records
        """
        results = []

        filters = self._coerce_field_types(filters or {})
        for record in self.list_all():
            matches = True
            for key, value in filters.items():
                if not hasattr(record, key) or getattr(record, key) != value:
                    matches = False
                    break
            if matches:
                results.append(record)

        if order_by:
            results = self._sort_items(results, order_by, sort_order)
        if offset:
            results = results[offset:]
        if limit:
            results = results[:limit]

        return results

    @ensure_store_exists
    def text_search(self, query: str, fields: list[str]) -> list[T]:
        """
        Search records with case-sensitive partial matching

        Args:
            query: Search string to look for
            fields: List of fields to search in

        Returns:
            List of matching records
        """
        results = []
        query = query

        for record in self.list_all():
            for field in fields:
                val = getattr(record, field, None)
                if val and query in val:
                    results.append(record)
                    break
        return results

    @ensure_store_exists
    def clear(self) -> None:
        """Clear all records in the store"""
        for file_path in self.item_type_dir.glob("*.yaml"):
            file_path.unlink()



from functools import wraps
from typing import Any, Callable

from syft_rds.client.rds_clients.base import RDSClientModule


def ensure_is_admin(func: Callable) -> Callable:
    """
    Decorator to ensure the user is an admin before executing a function.
    Admin status is determined by comparing the SyftBox client email with the configured host.
    """

    @wraps(func)
    def wrapper(self: RDSClientModule, *args: Any, **kwargs: Any) -> Callable:
        if not self.is_admin:
            raise PermissionError(
                f"You must be the datasite admin to perform this operation. "
                f"Your SyftBox email: '{self._syftbox_client.email}'. "
                f"Host email: '{self.config.host}'"
            )
        return func(self, *args, **kwargs)

    return wrapper



from syft_core import Client as SyftBoxClient

from syft_rds.server.app import APP_INFO_FILE, APP_NAME


def discover_rds_apps(syftbox_client: SyftBoxClient | None = None) -> list[str]:
    """Return all datasites that have the RDS app installed."""
    if syftbox_client is None:
        syftbox_client = SyftBoxClient.load()

    datasites_dir = syftbox_client.workspace.datasites

    datasites = []
    for dir in datasites_dir.iterdir():
        if not dir.is_dir():
            continue
        datasite_app_dir = syftbox_client.app_data(APP_NAME, datasite=dir.name)
        if (datasite_app_dir / APP_INFO_FILE).exists():
            datasites.append(dir.name)
    return datasites



from syft_rds.client.local_stores.base import CRUDLocalStore
from syft_rds.models import CustomFunction, CustomFunctionCreate, CustomFunctionUpdate


class CustomFunctionLocalStore(
    CRUDLocalStore[CustomFunction, CustomFunctionCreate, CustomFunctionUpdate]
):
    ITEM_TYPE = CustomFunction



from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
)

from syft_rpc import SyftResponse
from syft_rpc.rpc import BodyType

from syft_rds.client.connection import BlockingRPCConnection
from syft_rds.models import (
    Dataset,
    DatasetCreate,
    DatasetUpdate,
    GetAllRequest,
    GetOneRequest,
    ItemList,
    Job,
    JobCreate,
    JobUpdate,
    Runtime,
    RuntimeCreate,
    RuntimeUpdate,
    UserCode,
    UserCodeCreate,
    UserCodeUpdate,
)
from syft_rds.models.base import ItemBase, ItemBaseCreate, ItemBaseUpdate
from syft_rds.models.custom_function_models import (
    CustomFunction,
    CustomFunctionCreate,
    CustomFunctionUpdate,
)

if TYPE_CHECKING:
    from syft_rds.client.rds_client import RDSClientConfig


T = TypeVar("T", bound=ItemBase)
CreateT = TypeVar("CreateT", bound=ItemBaseCreate)
UpdateT = TypeVar("UpdateT", bound=ItemBaseUpdate)


class RPCClientModule:
    def __init__(self, config: "RDSClientConfig", connection: BlockingRPCConnection):
        self.config = config
        self.connection = connection

        self.prefix = f"syft://{self.config.host}/app_data/{self.config.app_name}/rpc"

    def _send(
        self, path: str, body: BodyType, expiry: Optional[Union[str, int]] = None
    ) -> SyftResponse:
        expiry = expiry or self.config.rpc_expiry
        if isinstance(expiry, int):
            expiry = f"{expiry}s"

        return self.connection.send(
            f"{self.prefix}/{path}",
            body,
            expiry=expiry,
            cache=False,
        )


class CRUDRPCClient(RPCClientModule, Generic[T, CreateT, UpdateT]):
    MODULE_NAME: ClassVar[str]
    ITEM_TYPE: ClassVar[type[T]]

    def register_client_id(self, item: T) -> T:
        if isinstance(item, ItemBase):
            item._register_client_id_recursive(self.config.uid)
        return item

    def create(self, item: CreateT) -> T:
        response = self._send(f"{self.MODULE_NAME}/create", item)
        response.raise_for_status()

        res = response.model(self.ITEM_TYPE)
        return self.register_client_id(res)

    def get_one(self, request: GetOneRequest) -> T:
        response = self._send(f"{self.MODULE_NAME}/get_one", request)
        response.raise_for_status()

        res = response.model(self.ITEM_TYPE)
        return self.register_client_id(res)

    def get_all(self, request: GetAllRequest) -> list[T]:
        response = self._send(f"{self.MODULE_NAME}/get_all", request)
        response.raise_for_status()

        item_list = response.model(ItemList[self.ITEM_TYPE])
        return [self.register_client_id(item) for item in item_list.items]

    def update(self, item: UpdateT) -> T:
        response = self._send(f"{self.MODULE_NAME}/update", item)
        response.raise_for_status()

        res = response.model(self.ITEM_TYPE)
        return self.register_client_id(res)


class DatasetRPCClient(CRUDRPCClient[Dataset, DatasetCreate, DatasetUpdate]):
    MODULE_NAME = "dataset"
    ITEM_TYPE = Dataset


class JobRPCClient(CRUDRPCClient[Job, JobCreate, JobUpdate]):
    MODULE_NAME = "job"
    ITEM_TYPE = Job


class RuntimeRPCClient(CRUDRPCClient[Runtime, RuntimeCreate, RuntimeUpdate]):
    MODULE_NAME = "runtime"
    ITEM_TYPE = Runtime


class UserCodeRPCClient(CRUDRPCClient[UserCode, UserCodeCreate, UserCodeUpdate]):
    MODULE_NAME = "user_code"
    ITEM_TYPE = UserCode


class CustomFunctionRPCClient(
    CRUDRPCClient[CustomFunction, CustomFunctionCreate, CustomFunctionUpdate]
):
    MODULE_NAME = "custom_function"
    ITEM_TYPE = CustomFunction


class RPCClient(RPCClientModule):
    def __init__(self, config: "RDSClientConfig", connection: BlockingRPCConnection):
        super().__init__(config, connection)

        self.job = JobRPCClient(self.config, self.connection)
        self.user_code = UserCodeRPCClient(self.config, self.connection)
        self.runtime = RuntimeRPCClient(self.config, self.connection)
        self.dataset = DatasetRPCClient(self.config, self.connection)
        self.custom_function = CustomFunctionRPCClient(self.config, self.connection)

        # Create lookup table for type-based access
        self._type_map = {
            Job: self.job,
            UserCode: self.user_code,
            Runtime: self.runtime,
            Dataset: self.dataset,
            CustomFunction: self.custom_function,
        }

    def for_type(
        self, type_: Type[T]
    ) -> CRUDRPCClient[T, ItemBaseCreate, ItemBaseUpdate]:
        if type_ not in self._type_map:
            raise ValueError(f"No client registered for type {type_}")
        return self._type_map[type_]

    def health(self, expiry: Optional[Union[str, int]] = None) -> dict:
        response: SyftResponse = self._send("/health", body=None, expiry=expiry)
        response.raise_for_status()

        return response.json()



from pathlib import Path
from typing import Optional

import typer
from loguru import logger
from syft_core import Client as SyftBoxClient
from syft_core import SyftClientConfig

from syft_rds.server.app import create_app

app = typer.Typer(
    name="syft-rds",
    help="Syft RDS CLI",
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@app.command()
def server(
    syftbox_config: Optional[Path] = typer.Option(None),
) -> None:
    """Start the RDS server."""
    syftbox_client = SyftBoxClient.load(filepath=syftbox_config)
    typer.echo(f"SyftBox client loaded from {syftbox_client.config_path}")
    rds_app = create_app(client=syftbox_client)
    rds_app.run_forever()


@app.command()
def init_test_datasite(
    email: str = typer.Option(..., help="Email address for the client"),
    client_url: str = typer.Option(
        "http://testserver:5000",
        help="URL of the SyftBox client server. Not used when testing locally.",
    ),
    data_dir: Path = typer.Option(..., help="Directory for client data"),
    config_path: Path = typer.Option(..., help="Path to save the config file"),
    overwrite: bool = typer.Option(False, help="Overwrite existing config if present"),
) -> None:
    """
    Initialize a new SyftBox client config for a datasite.
    """
    config_path = config_path.resolve()
    data_dir = data_dir.resolve()

    if config_path.exists() and not overwrite:
        logger.warning(
            f"Config file {config_path} already exists. Use --overwrite to replace."
        )
        typer.echo(f"Config file {config_path} already exists. No action taken.")
        raise typer.Exit(code=0)

    config = SyftClientConfig(
        email=email,
        client_url=client_url,
        path=config_path,
        data_dir=data_dir,
    )
    config.save()
    typer.echo(f"Created new SyftBox client config at {config_path}")


@app.callback()
def callback():
    # Empty command to enable subcommands
    pass


def main():
    app()


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Optional

from syft_core import Client as SyftBoxClient
from syft_core import SyftBoxURL
from syft_core.permissions import (
    get_computed_permission,
    ComputedPermission,
    PermissionType,
)
from syft_core.exceptions import SyftBoxException
from syft_core.types import RelativePath
from syft_event import SyftEvents
from syft_event.deps import func_args_from_request
from syft_rpc import SyftRequest, SyftResponse, rpc
from syft_rpc.protocol import SyftMethod, SyftStatus
from syft_rpc.rpc import BodyType


class BlockingRPCConnection(ABC):
    def __init__(
        self,
        sender_client: SyftBoxClient,
        default_expiry: str = "15m",
    ):
        self.sender_client = sender_client
        self.default_expiry = default_expiry

    @abstractmethod
    def send(
        self,
        url: str,
        body: BodyType,
        headers: Optional[dict] = None,
        expiry: Optional[str] = None,
        cache: bool = False,
    ) -> SyftResponse:
        raise NotImplementedError()

    def _serialize(self, body: BodyType) -> str:
        # NOTE to enable partial BaseModel updates, always exclude unset fields when serializing
        # exclude_unset will only affect the serialization of Pydantic models
        return rpc.serialize(body, exclude_unset=True)


class FileSyncRPCConnection(BlockingRPCConnection):
    def send(
        self,
        url: str,
        body: BodyType,
        headers: Optional[dict] = None,
        expiry: Optional[str] = None,
        cache: bool = False,
    ) -> SyftResponse:
        headers = None

        body = self._serialize(body)
        future = rpc.send(
            url=url,
            body=body,
            headers=headers,
            expiry=expiry,
            cache=cache,
            client=self.sender_client,
        )

        timeout_seconds = float(rpc.parse_duration(expiry).seconds)
        response = future.wait(timeout=timeout_seconds)
        return response


def check_permission(
    client: SyftBoxClient,
    path: str,
    permission: PermissionType,
) -> bool:
    """
    Check if the client has the permission to access the url.
    If the client does not have the permission, we will raise an error *in future*,
    but for now we will just log a warning.
    """
    try:
        relative_path = RelativePath(path)
    except ValueError:
        path = path.relative_to(client.workspace.datasites)
        relative_path = RelativePath(path)
    sender_permission: ComputedPermission = get_computed_permission(
        client=client, path=relative_path
    )
    has_permission = sender_permission.has_permission(permission)
    if not has_permission:
        raise SyftBoxException(
            f"User {client.email} does not have {permission} permission for {path}"
        )

    return has_permission


class MockRPCConnection(BlockingRPCConnection):
    app: SyftEvents
    sender_client: SyftBoxClient

    def __init__(self, sender_client: SyftBoxClient, app: SyftEvents):
        self.app = app
        self.app.init()
        super().__init__(sender_client)

    @property
    def receiver_client(self) -> SyftBoxClient:
        return self.app.client

    def _build_request(
        self,
        url: str,
        body: BodyType,
        headers: Optional[dict] = None,
        expiry: Optional[str] = None,
    ) -> SyftRequest:
        headers = None

        expiry_time = datetime.now(timezone.utc) + rpc.parse_duration(expiry)
        return SyftRequest(
            sender=self.sender_client.email,
            method=SyftMethod.GET,
            url=url if isinstance(url, SyftBoxURL) else SyftBoxURL(url),
            headers=headers or {},
            body=rpc.serialize(body),
            expires=expiry_time,
        )

    def _build_response(
        self,
        request: SyftRequest,
        response_body: BodyType,
        status_code: SyftStatus = SyftStatus.SYFT_200_OK,
    ) -> SyftResponse:
        return SyftResponse(
            id=request.id,
            sender=self.receiver_client.email,
            url=request.url,
            headers={},
            body=rpc.serialize(response_body),
            expires=request.expires,
            status_code=status_code,
        )

    def send(
        self,
        url: str,
        body: BodyType,
        headers: Optional[dict] = None,
        expiry: Optional[str] = None,
        cache: bool = False,
    ) -> SyftResponse:
        if cache:
            raise NotImplementedError("Cache not implemented for MockRPCConnection")

        # NOTE to match the FileSyncRPCConnection.send implementation, we self._serialize the body here with our custom serde options
        # in rpc.send the body will be serialized again which will be a no-op when building the request. This is a no-op on already serialized data.
        body = self._serialize(body)
        syft_request = self._build_request(url, body, headers, expiry)
        syft_url = SyftBoxURL(url)

        req_path = (
            syft_url.to_local_path(self.sender_client.workspace.datasites)
            / f"{syft_request.id}.request"
        )

        req_path = req_path.relative_to(self.sender_client.workspace.datasites)

        # check_permission(self.sender_client, req_path, PermissionType.WRITE)
        # check_permission(self.receiver_client, req_path, PermissionType.READ)

        receiver_local_path = syft_url.to_local_path(
            self.receiver_client.workspace.datasites
        )
        handler = self.app.get_handler(receiver_local_path)
        if handler is None:
            raise ValueError(f"No handler found for: {receiver_local_path}")
        kwargs = func_args_from_request(handler, syft_request, self.app)

        response_body = handler(**kwargs)
        return self._build_response(syft_request, response_body)


def get_connection(
    sender_client: SyftBoxClient,
    app: SyftEvents,
    mock: bool = False,
) -> BlockingRPCConnection:
    if mock:
        return MockRPCConnection(sender_client=sender_client, app=app)
    else:
        return FileSyncRPCConnection(sender_client=sender_client)
